""" Test to ensure that the code is working correctly.
Runs all metrics on 14 trackers for the MOT Challenge MOT17 benchmark.
"""


import sys
import os
import numpy as np
from multiprocessing import freeze_support

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import trackeval  # noqa: E402

# Fixes multiprocessing on windows, does nothing otherwise
if __name__ == '__main__':
    freeze_support()

eval_config = {'USE_PARALLEL': False,
               'NUM_PARALLEL_CORES': 8,
               }
evaluator = trackeval.Evaluator(eval_config)
metrics_list = [trackeval.metrics.HOTA(), trackeval.metrics.CLEAR(), trackeval.metrics.Identity()]
test_data_loc = os.path.join(os.path.dirname(__file__), '..', 'data', 'tests', 'mot_challenge', 'MOT17-train')
trackers = [
    'DPMOT',
    'GNNMatch',
    'IA',
    'ISE_MOT17R',
    'Lif_T',
    'Lif_TsimInt',
    'LPC_MOT',
    'MAT',
    'MIFTv2',
    'MPNTrack',
    'SSAT',
    'TracktorCorr',
    'Tracktorv2',
    'UnsupTrack',
]

for tracker in trackers:
    # Run code on tracker
    dataset_config = {'TRACKERS_TO_EVAL': [tracker],
                      'BENCHMARK': 'MOT17'}
    dataset_list = [trackeval.datasets.MotChallenge2DBox(dataset_config)]
    raw_results, messages = evaluator.evaluate(dataset_list, metrics_list)

    results = {seq: raw_results['MotChallenge2DBox'][tracker][seq]['pedestrian'] for seq in
               raw_results['MotChallenge2DBox'][tracker].keys()}
    current_metrics_list = metrics_list + [trackeval.metrics.Count()]
    metric_names = trackeval.utils.validate_metrics_list(current_metrics_list)

    # Load expected results:
    test_data = trackeval.utils.load_detail(os.path.join(test_data_loc, tracker, 'pedestrian_detailed.csv'))
    assert len(test_data.keys()) == 22, len(test_data.keys())

    # Do checks
    for seq in test_data.keys():
        assert len(test_data[seq].keys()) > 250, len(test_data[seq].keys())

        details = []
        for metric, metric_name in zip(current_metrics_list, metric_names):
            table_res = {seq_key: seq_value[metric_name] for seq_key, seq_value in results.items()}
            details.append(metric.detailed_results(table_res))
        res_fields = sum([list(s['COMBINED_SEQ'].keys()) for s in details], [])
        res_values = sum([list(s[seq].values()) for s in details], [])
        res_dict = dict(zip(res_fields, res_values))

        for field in test_data[seq].keys():
            if not np.isclose(res_dict[field], test_data[seq][field]):
                print(tracker, seq, res_dict[field], test_data[seq][field], field)
                raise AssertionError

    print('Tracker %s tests passed' % tracker)
print('All tests passed')





""" run_person_path_22.py

Run example:
python3 run_person_path_22.py \
    --BENCHMARK person_path_22 \
    --SPLIT_TO_EVAL test \
    --TRACKERS_TO_EVAL custom_tracker \
    --METRICS HOTA CLEAR Identity VACE \
    --USE_PARALLEL True \
    --NUM_PARALLEL_CORES 12 \
    --TRACKERS_FOLDER $TRACKER_FOLDER

where $TRACKER_FOLDER is the path of the folder containing the tracker predictions in MOTChallenge format.
In particular, $TRACKER_FOLDER is expected to have the following structure:
$TRACKER_FOLDER/
    person_path_22-test/
        custom_tracker/
            data/
                uid_vid_00008.mp4.txt
                uid_vid_00009.mp4.txt
                [...]
                uid_vid_00235.mp4.txt

Each text file contains the tracker predictions in MOTChallenge format for a given video.


Command Line Arguments: Defaults, # Comments
    Eval arguments:
        'USE_PARALLEL': False,
        'NUM_PARALLEL_CORES': 8,
        'BREAK_ON_ERROR': True,
        'PRINT_RESULTS': True,
        'PRINT_ONLY_COMBINED': False,
        'PRINT_CONFIG': True,
        'TIME_PROGRESS': True,
        'OUTPUT_SUMMARY': True,
        'OUTPUT_DETAILED': True,
        'PLOT_CURVES': True,
    Dataset arguments:
        'GT_FOLDER': os.path.join(code_path, 'data/gt/person_path_22/'),  # Location of GT data
        'TRACKERS_FOLDER': os.path.join(code_path, 'data/trackers/person_path_22/'),  # Trackers location
        'OUTPUT_FOLDER': None,  # Where to save eval results (if None, same as TRACKERS_FOLDER)
        'TRACKERS_TO_EVAL': None,  # Filenames of trackers to eval (if None, all in folder)
        'CLASSES_TO_EVAL': ['pedestrian'],  # Valid: ['pedestrian']
        'BENCHMARK': 'person_path_22',  # Valid: 'person_path_22'
        'SPLIT_TO_EVAL': 'train',  # Valid: 'train', 'test', 'all'
        'INPUT_AS_ZIP': False,  # Whether tracker input files are zipped
        'PRINT_CONFIG': True,  # Whether to print current config
        'DO_PREPROC': True,  # Whether to perform preprocessing
        'TRACKER_SUB_FOLDER': 'data',  # Tracker files are in TRACKER_FOLDER/tracker_name/TRACKER_SUB_FOLDER
        'OUTPUT_SUB_FOLDER': '',  # Output files are saved in OUTPUT_FOLDER/tracker_name/OUTPUT_SUB_FOLDER
    Metric arguments:
        'METRICS': ['HOTA', 'CLEAR', 'Identity', 'VACE']
"""

import sys
import os
import argparse
from multiprocessing import freeze_support

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import trackeval  # noqa: E402

if __name__ == '__main__':
    freeze_support()

    # Command line interface:
    default_eval_config = trackeval.Evaluator.get_default_eval_config()
    default_eval_config['DISPLAY_LESS_PROGRESS'] = False
    default_dataset_config = trackeval.datasets.PersonPath22.get_default_dataset_config()
    default_metrics_config = {'METRICS': ['HOTA', 'CLEAR', 'Identity'], 'THRESHOLD': 0.5}
    config = {**default_eval_config, **default_dataset_config, **default_metrics_config}  # Merge default configs
    parser = argparse.ArgumentParser()
    for setting in config.keys():
        if type(config[setting]) == list or type(config[setting]) == type(None):
            parser.add_argument("--" + setting, nargs='+')
        else:
            parser.add_argument("--" + setting)
    args = parser.parse_args().__dict__
    for setting in args.keys():
        if args[setting] is not None:
            if type(config[setting]) == type(True):
                if args[setting] == 'True':
                    x = True
                elif args[setting] == 'False':
                    x = False
                else:
                    raise Exception('Command line parameter ' + setting + 'must be True or False')
            elif type(config[setting]) == type(1):
                x = int(args[setting])
            elif type(args[setting]) == type(None):
                x = None
            elif setting == 'SEQ_INFO':
                x = dict(zip(args[setting], [None]*len(args[setting])))
            else:
                x = args[setting]
            config[setting] = x
    eval_config = {k: v for k, v in config.items() if k in default_eval_config.keys()}
    dataset_config = {k: v for k, v in config.items() if k in default_dataset_config.keys()}
    metrics_config = {k: v for k, v in config.items() if k in default_metrics_config.keys()}

    # Run code
    evaluator = trackeval.Evaluator(eval_config)
    dataset_list = [trackeval.datasets.PersonPath22(dataset_config)]
    metrics_list = []
    for metric in [trackeval.metrics.HOTA, trackeval.metrics.CLEAR, trackeval.metrics.Identity, trackeval.metrics.VACE]:
        if metric.get_name() in metrics_config['METRICS']:
            metrics_list.append(metric(metrics_config))
    if len(metrics_list) == 0:
        raise Exception('No metrics selected for evaluation')
    evaluator.evaluate(dataset_list, metrics_list)



import numpy as np
from scipy.optimize import linear_sum_assignment
from ._base_metric import _BaseMetric
from .. import _timing


class VACE(_BaseMetric):
    """Class which implements the VACE metrics.

    The metrics are described in:
    Manohar et al. (2006) "Performance Evaluation of Object Detection and Tracking in Video"
    https://link.springer.com/chapter/10.1007/11612704_16

    This implementation uses the "relaxed" variant of the metrics,
    where an overlap threshold is applied in each frame.
    """

    def __init__(self, config=None):
        super().__init__()
        self.integer_fields = ['VACE_IDs', 'VACE_GT_IDs', 'num_non_empty_timesteps']
        self.float_fields = ['STDA', 'ATA', 'FDA', 'SFDA']
        self.fields = self.integer_fields + self.float_fields
        self.summary_fields = ['SFDA', 'ATA']

        # Fields that are accumulated over multiple videos.
        self._additive_fields = self.integer_fields + ['STDA', 'FDA']

        self.threshold = 0.5

    @_timing.time
    def eval_sequence(self, data):
        """Calculates VACE metrics for one sequence.

        Depends on the fields:
            data['num_gt_ids']
            data['num_tracker_ids']
            data['gt_ids']
            data['tracker_ids']
            data['similarity_scores']
        """
        res = {}

        # Obtain Average Tracking Accuracy (ATA) using track correspondence.
        # Obtain counts necessary to compute temporal IOU.
        # Assume that integer counts can be represented exactly as floats.
        potential_matches_count = np.zeros((data['num_gt_ids'], data['num_tracker_ids']))
        gt_id_count = np.zeros(data['num_gt_ids'])
        tracker_id_count = np.zeros(data['num_tracker_ids'])
        both_present_count = np.zeros((data['num_gt_ids'], data['num_tracker_ids']))
        for t, (gt_ids_t, tracker_ids_t) in enumerate(zip(data['gt_ids'], data['tracker_ids'])):
            # Count the number of frames in which two tracks satisfy the overlap criterion.
            matches_mask = np.greater_equal(data['similarity_scores'][t], self.threshold)
            match_idx_gt, match_idx_tracker = np.nonzero(matches_mask)
            potential_matches_count[gt_ids_t[match_idx_gt], tracker_ids_t[match_idx_tracker]] += 1
            # Count the number of frames in which the tracks are present.
            gt_id_count[gt_ids_t] += 1
            tracker_id_count[tracker_ids_t] += 1
            both_present_count[gt_ids_t[:, np.newaxis], tracker_ids_t[np.newaxis, :]] += 1
        # Number of frames in which either track is present (union of the two sets of frames).
        union_count = (gt_id_count[:, np.newaxis]
                       + tracker_id_count[np.newaxis, :]
                       - both_present_count)
        # The denominator should always be non-zero if all tracks are non-empty.
        with np.errstate(divide='raise', invalid='raise'):
            temporal_iou = potential_matches_count / union_count
        # Find assignment that maximizes temporal IOU.
        match_rows, match_cols = linear_sum_assignment(-temporal_iou)
        res['STDA'] = temporal_iou[match_rows, match_cols].sum()
        res['VACE_IDs'] = data['num_tracker_ids']
        res['VACE_GT_IDs'] = data['num_gt_ids']

        # Obtain Frame Detection Accuracy (FDA) using per-frame correspondence.
        non_empty_count = 0
        fda = 0
        for t, (gt_ids_t, tracker_ids_t) in enumerate(zip(data['gt_ids'], data['tracker_ids'])):
            n_g = len(gt_ids_t)
            n_d = len(tracker_ids_t)
            if not (n_g or n_d):
                continue
            # n_g > 0 or n_d > 0
            non_empty_count += 1
            if not (n_g and n_d):
                continue
            # n_g > 0 and n_d > 0
            spatial_overlap = data['similarity_scores'][t]
            match_rows, match_cols = linear_sum_assignment(-spatial_overlap)
            overlap_ratio = spatial_overlap[match_rows, match_cols].sum()
            fda += overlap_ratio / (0.5 * (n_g + n_d))
        res['FDA'] = fda
        res['num_non_empty_timesteps'] = non_empty_count

        res.update(self._compute_final_fields(res))
        return res

    def combine_classes_class_averaged(self, all_res, ignore_empty_classes=True):
        """Combines metrics across all classes by averaging over the class values.
        If 'ignore_empty_classes' is True, then it only sums over classes with at least one gt or predicted detection.
        """
        res = {}
        for field in self.fields:
            if ignore_empty_classes:
                res[field] = np.mean([v[field] for v in all_res.values()
                                  if v['VACE_GT_IDs'] > 0 or v['VACE_IDs'] > 0], axis=0)
            else:
                res[field] = np.mean([v[field] for v in all_res.values()], axis=0)
        return res

    def combine_classes_det_averaged(self, all_res):
        """Combines metrics across all classes by averaging over the detection values"""
        res = {}
        for field in self._additive_fields:
            res[field] = _BaseMetric._combine_sum(all_res, field)
        res = self._compute_final_fields(res)
        return res

    def combine_sequences(self, all_res):
        """Combines metrics across all sequences"""
        res = {}
        for header in self._additive_fields:
            res[header] = _BaseMetric._combine_sum(all_res, header)
        res.update(self._compute_final_fields(res))
        return res

    @staticmethod
    def _compute_final_fields(additive):
        final = {}
        with np.errstate(invalid='ignore'):  # Permit nan results.
            final['ATA'] = (additive['STDA'] /
                            (0.5 * (additive['VACE_IDs'] + additive['VACE_GT_IDs'])))
            final['SFDA'] = additive['FDA'] / additive['num_non_empty_timesteps']
        return final



""" run_davis.py

Run example:
run_davis.py --USE_PARALLEL False --METRICS HOTA --TRACKERS_TO_EVAL ags

Command Line Arguments: Defaults, # Comments
    Eval arguments:
        'USE_PARALLEL': False,
        'NUM_PARALLEL_CORES': 8,
        'BREAK_ON_ERROR': True,
        'PRINT_RESULTS': True,
        'PRINT_ONLY_COMBINED': False,
        'PRINT_CONFIG': True,
        'TIME_PROGRESS': True,
        'OUTPUT_SUMMARY': True,
        'OUTPUT_DETAILED': True,
        'PLOT_CURVES': True,
    Dataset arguments:
    '   'GT_FOLDER': os.path.join(code_path, 'data/gt/davis/'),  # Location of GT data
        'TRACKERS_FOLDER': os.path.join(code_path, 'data/trackers/davis/davis_val'),  # Trackers location
        'OUTPUT_FOLDER': None,  # Where to save eval results (if None, same as TRACKERS_FOLDER)
        'TRACKERS_TO_EVAL': None,  # Filenames of trackers to eval (if None, all in folder)
        'SPLIT_TO_EVAL': 'val',  # Valid: 'val', 'train'
        'PRINT_CONFIG': True,  # Whether to print current config
        'TRACKER_SUB_FOLDER': 'data',  # Tracker files are in TRACKER_FOLDER/tracker_name/TRACKER_SUB_FOLDER
        'OUTPUT_SUB_FOLDER': '',  # Output files are saved in OUTPUT_FOLDER/tracker_name/OUTPUT_SUB_FOLDER
        'TRACKER_DISPLAY_NAMES': None,  # Names of trackers to display, if None: TRACKERS_TO_EVAL
        'SEQMAP_FOLDER': None,  # Where seqmaps are found (if None, GT_FOLDER/ImageSets/2017)
        'SEQMAP_FILE': None,  # Directly specify seqmap file (if none use seqmap_folder/split-to-eval.txt)
        'SEQ_INFO': None,  # If not None, directly specify sequences to eval and their number of timesteps
        'GT_LOC_FORMAT': '{gt_folder}/Annotations_unsupervised/480p/{seq}',
        # '{gt_folder}/Annotations_unsupervised/480p/{seq}'
        'MAX_DETECTIONS': 0  # Maximum number of allowed detections per sequence (0 for no threshold)
    Metric arguments:
        'METRICS': ['HOTA', 'CLEAR', 'Identity', 'JAndF']
"""

import sys
import os
import argparse
from multiprocessing import freeze_support

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import trackeval  # noqa: E402

if __name__ == '__main__':
    freeze_support()

    # Command line interface:
    default_eval_config = trackeval.Evaluator.get_default_eval_config()
    default_dataset_config = trackeval.datasets.DAVIS.get_default_dataset_config()
    default_metrics_config = {'METRICS': ['HOTA', 'CLEAR', 'Identity', 'JAndF']}
    config = {**default_eval_config, **default_dataset_config, **default_metrics_config}  # Merge default configs
    parser = argparse.ArgumentParser()
    for setting in config.keys():
        if type(config[setting]) == list or type(config[setting]) == type(None):
            parser.add_argument("--" + setting, nargs='+')
        else:
            parser.add_argument("--" + setting)
    args = parser.parse_args().__dict__
    for setting in args.keys():
        if args[setting] is not None:
            if type(config[setting]) == type(True):
                if args[setting] == 'True':
                    x = True
                elif args[setting] == 'False':
                    x = False
                else:
                    raise Exception('Command line parameter ' + setting + 'must be True or False')
            elif type(config[setting]) == type(1):
                x = int(args[setting])
            elif type(args[setting]) == type(None):
                x = None
            else:
                x = args[setting]
            config[setting] = x
    eval_config = {k: v for k, v in config.items() if k in default_eval_config.keys()}
    dataset_config = {k: v for k, v in config.items() if k in default_dataset_config.keys()}
    metrics_config = {k: v for k, v in config.items() if k in default_metrics_config.keys()}

    # Run code
    evaluator = trackeval.Evaluator(eval_config)
    dataset_list = [trackeval.datasets.DAVIS(dataset_config)]
    metrics_list = []
    for metric in [trackeval.metrics.HOTA, trackeval.metrics.CLEAR, trackeval.metrics.Identity, trackeval.metrics.JAndF]:
        if metric.get_name() in metrics_config['METRICS']:
            metrics_list.append(metric())
    if len(metrics_list) == 0:
        raise Exception('No metrics selected for evaluation')
    evaluator.evaluate(dataset_list, metrics_list)



import os
import numpy as np
from scipy.optimize import linear_sum_assignment
from ._base_metric import _BaseMetric
from .. import _timing


class HOTA(_BaseMetric):
    """Class which implements the HOTA metrics.
    See: https://link.springer.com/article/10.1007/s11263-020-01375-2
    """

    def __init__(self, config=None):
        super().__init__()
        self.plottable = True
        self.array_labels = np.arange(0.05, 0.99, 0.05)
        self.integer_array_fields = ['HOTA_TP', 'HOTA_FN', 'HOTA_FP']
        self.float_array_fields = ['HOTA', 'DetA', 'AssA', 'DetRe', 'DetPr', 'AssRe', 'AssPr', 'LocA', 'OWTA']
        self.float_fields = ['HOTA(0)', 'LocA(0)', 'HOTALocA(0)']
        self.fields = self.float_array_fields + self.integer_array_fields + self.float_fields
        self.summary_fields = self.float_array_fields + self.float_fields

    @_timing.time
    def eval_sequence(self, data):
        """Calculates the HOTA metrics for one sequence"""

        # Initialise results
        res = {}
        for field in self.float_array_fields + self.integer_array_fields:
            res[field] = np.zeros((len(self.array_labels)), dtype=float)
        for field in self.float_fields:
            res[field] = 0

        # Return result quickly if tracker or gt sequence is empty
        if data['num_tracker_dets'] == 0:
            res['HOTA_FN'] = data['num_gt_dets'] * np.ones((len(self.array_labels)), dtype=float)
            res['LocA'] = np.ones((len(self.array_labels)), dtype=float)
            res['LocA(0)'] = 1.0
            return res
        if data['num_gt_dets'] == 0:
            res['HOTA_FP'] = data['num_tracker_dets'] * np.ones((len(self.array_labels)), dtype=float)
            res['LocA'] = np.ones((len(self.array_labels)), dtype=float)
            res['LocA(0)'] = 1.0
            return res

        # Variables counting global association
        potential_matches_count = np.zeros((data['num_gt_ids'], data['num_tracker_ids']))
        gt_id_count = np.zeros((data['num_gt_ids'], 1))
        tracker_id_count = np.zeros((1, data['num_tracker_ids']))

        # First loop through each timestep and accumulate global track information.
        for t, (gt_ids_t, tracker_ids_t) in enumerate(zip(data['gt_ids'], data['tracker_ids'])):
            # Count the potential matches between ids in each timestep
            # These are normalised, weighted by the match similarity.
            similarity = data['similarity_scores'][t]
            sim_iou_denom = similarity.sum(0)[np.newaxis, :] + similarity.sum(1)[:, np.newaxis] - similarity
            sim_iou = np.zeros_like(similarity)
            sim_iou_mask = sim_iou_denom > 0 + np.finfo('float').eps
            sim_iou[sim_iou_mask] = similarity[sim_iou_mask] / sim_iou_denom[sim_iou_mask]
            potential_matches_count[gt_ids_t[:, np.newaxis], tracker_ids_t[np.newaxis, :]] += sim_iou

            # Calculate the total number of dets for each gt_id and tracker_id.
            gt_id_count[gt_ids_t] += 1
            tracker_id_count[0, tracker_ids_t] += 1

        # Calculate overall jaccard alignment score (before unique matching) between IDs
        global_alignment_score = potential_matches_count / (gt_id_count + tracker_id_count - potential_matches_count)
        matches_counts = [np.zeros_like(potential_matches_count) for _ in self.array_labels]

        # Calculate scores for each timestep
        for t, (gt_ids_t, tracker_ids_t) in enumerate(zip(data['gt_ids'], data['tracker_ids'])):
            # Deal with the case that there are no gt_det/tracker_det in a timestep.
            if len(gt_ids_t) == 0:
                for a, alpha in enumerate(self.array_labels):
                    res['HOTA_FP'][a] += len(tracker_ids_t)
                continue
            if len(tracker_ids_t) == 0:
                for a, alpha in enumerate(self.array_labels):
                    res['HOTA_FN'][a] += len(gt_ids_t)
                continue

            # Get matching scores between pairs of dets for optimizing HOTA
            similarity = data['similarity_scores'][t]
            score_mat = global_alignment_score[gt_ids_t[:, np.newaxis], tracker_ids_t[np.newaxis, :]] * similarity

            # Hungarian algorithm to find best matches
            match_rows, match_cols = linear_sum_assignment(-score_mat)

            # Calculate and accumulate basic statistics
            for a, alpha in enumerate(self.array_labels):
                actually_matched_mask = similarity[match_rows, match_cols] >= alpha - np.finfo('float').eps
                alpha_match_rows = match_rows[actually_matched_mask]
                alpha_match_cols = match_cols[actually_matched_mask]
                num_matches = len(alpha_match_rows)
                res['HOTA_TP'][a] += num_matches
                res['HOTA_FN'][a] += len(gt_ids_t) - num_matches
                res['HOTA_FP'][a] += len(tracker_ids_t) - num_matches
                if num_matches > 0:
                    res['LocA'][a] += sum(similarity[alpha_match_rows, alpha_match_cols])
                    matches_counts[a][gt_ids_t[alpha_match_rows], tracker_ids_t[alpha_match_cols]] += 1

        # Calculate association scores (AssA, AssRe, AssPr) for the alpha value.
        # First calculate scores per gt_id/tracker_id combo and then average over the number of detections.
        for a, alpha in enumerate(self.array_labels):
            matches_count = matches_counts[a]
            ass_a = matches_count / np.maximum(1, gt_id_count + tracker_id_count - matches_count)
            res['AssA'][a] = np.sum(matches_count * ass_a) / np.maximum(1, res['HOTA_TP'][a])
            ass_re = matches_count / np.maximum(1, gt_id_count)
            res['AssRe'][a] = np.sum(matches_count * ass_re) / np.maximum(1, res['HOTA_TP'][a])
            ass_pr = matches_count / np.maximum(1, tracker_id_count)
            res['AssPr'][a] = np.sum(matches_count * ass_pr) / np.maximum(1, res['HOTA_TP'][a])

        # Calculate final scores
        res['LocA'] = np.maximum(1e-10, res['LocA']) / np.maximum(1e-10, res['HOTA_TP'])
        res = self._compute_final_fields(res)
        return res

    def combine_sequences(self, all_res):
        """Combines metrics across all sequences"""
        res = {}
        for field in self.integer_array_fields:
            res[field] = self._combine_sum(all_res, field)
        for field in ['AssRe', 'AssPr', 'AssA']:
            res[field] = self._combine_weighted_av(all_res, field, res, weight_field='HOTA_TP')
        loca_weighted_sum = sum([all_res[k]['LocA'] * all_res[k]['HOTA_TP'] for k in all_res.keys()])
        res['LocA'] = np.maximum(1e-10, loca_weighted_sum) / np.maximum(1e-10, res['HOTA_TP'])
        res = self._compute_final_fields(res)
        return res

    def combine_classes_class_averaged(self, all_res, ignore_empty_classes=False):
        """Combines metrics across all classes by averaging over the class values.
        If 'ignore_empty_classes' is True, then it only sums over classes with at least one gt or predicted detection.
        """
        res = {}
        for field in self.integer_array_fields:
            if ignore_empty_classes:
                res[field] = self._combine_sum(
                    {k: v for k, v in all_res.items()
                     if (v['HOTA_TP'] + v['HOTA_FN'] + v['HOTA_FP'] > 0 + np.finfo('float').eps).any()}, field)
            else:
                res[field] = self._combine_sum({k: v for k, v in all_res.items()}, field)

        for field in self.float_fields + self.float_array_fields:
            if ignore_empty_classes:
                res[field] = np.mean([v[field] for v in all_res.values() if
                                      (v['HOTA_TP'] + v['HOTA_FN'] + v['HOTA_FP'] > 0 + np.finfo('float').eps).any()],
                                     axis=0)
            else:
                res[field] = np.mean([v[field] for v in all_res.values()], axis=0)
        return res

    def combine_classes_det_averaged(self, all_res):
        """Combines metrics across all classes by averaging over the detection values"""
        res = {}
        for field in self.integer_array_fields:
            res[field] = self._combine_sum(all_res, field)
        for field in ['AssRe', 'AssPr', 'AssA']:
            res[field] = self._combine_weighted_av(all_res, field, res, weight_field='HOTA_TP')
        loca_weighted_sum = sum([all_res[k]['LocA'] * all_res[k]['HOTA_TP'] for k in all_res.keys()])
        res['LocA'] = np.maximum(1e-10, loca_weighted_sum) / np.maximum(1e-10, res['HOTA_TP'])
        res = self._compute_final_fields(res)
        return res

    @staticmethod
    def _compute_final_fields(res):
        """Calculate sub-metric ('field') values which only depend on other sub-metric values.
        This function is used both for both per-sequence calculation, and in combining values across sequences.
        """
        res['DetRe'] = res['HOTA_TP'] / np.maximum(1, res['HOTA_TP'] + res['HOTA_FN'])
        res['DetPr'] = res['HOTA_TP'] / np.maximum(1, res['HOTA_TP'] + res['HOTA_FP'])
        res['DetA'] = res['HOTA_TP'] / np.maximum(1, res['HOTA_TP'] + res['HOTA_FN'] + res['HOTA_FP'])
        res['HOTA'] = np.sqrt(res['DetA'] * res['AssA'])
        res['OWTA'] = np.sqrt(res['DetRe'] * res['AssA'])

        res['HOTA(0)'] = res['HOTA'][0]
        res['LocA(0)'] = res['LocA'][0]
        res['HOTALocA(0)'] = res['HOTA(0)']*res['LocA(0)']
        return res

    def plot_single_tracker_results(self, table_res, tracker, cls, output_folder):
        """Create plot of results"""

        # Only loaded when run to reduce minimum requirements
        from matplotlib import pyplot as plt

        res = table_res['COMBINED_SEQ']
        styles_to_plot = ['r', 'b', 'g', 'b--', 'b:', 'g--', 'g:', 'm']
        for name, style in zip(self.float_array_fields, styles_to_plot):
            plt.plot(self.array_labels, res[name], style)
        plt.xlabel('alpha')
        plt.ylabel('score')
        plt.title(tracker + ' - ' + cls)
        plt.axis([0, 1, 0, 1])
        legend = []
        for name in self.float_array_fields:
            legend += [name + ' (' + str(np.round(np.mean(res[name]), 2)) + ')']
        plt.legend(legend, loc='lower left')
        out_file = os.path.join(output_folder, cls + '_plot.pdf')
        os.makedirs(os.path.dirname(out_file), exist_ok=True)
        plt.savefig(out_file)
        plt.savefig(out_file.replace('.pdf', '.png'))
        plt.clf()



#!/usr/bin/env python3
"""
BoxMOT包构建脚本

这个脚本自动化了包的构建、测试和发布流程。

使用方法:
    python build_package.py --help
    python build_package.py build
    python build_package.py test
    python build_package.py publish --test
    python build_package.py publish --production
"""

import argparse
import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import List, Optional


class PackageBuilder:
    """包构建器类"""
    
    def __init__(self, project_root: Path):
        """初始化构建器
        
        Args:
            project_root: 项目根目录路径
        """
        self.project_root = project_root
        self.dist_dir = project_root / "dist"
        self.build_dir = project_root / "build"
        
    def clean(self) -> bool:
        """清理构建文件
        
        Returns:
            bool: 清理是否成功
        """
        print("🧹 清理构建文件...")
        
        # 要清理的目录和文件模式
        clean_targets = [
            self.dist_dir,
            self.build_dir,
            self.project_root / "*.egg-info",
            self.project_root / "__pycache__",
        ]
        
        try:
            for target in clean_targets:
                if target.exists():
                    if target.is_dir():
                        shutil.rmtree(target)
                        print(f"  ✓ 删除目录: {target}")
                    else:
                        target.unlink()
                        print(f"  ✓ 删除文件: {target}")
                        
            # 清理Python缓存
            for cache_dir in self.project_root.rglob("__pycache__"):
                shutil.rmtree(cache_dir)
                print(f"  ✓ 删除缓存: {cache_dir}")
                
            print("✅ 清理完成")
            return True
            
        except Exception as e:
            print(f"❌ 清理失败: {e}")
            return False
    
    def build(self) -> bool:
        """构建包
        
        Returns:
            bool: 构建是否成功
        """
        print("🏗️  开始构建包...")
        
        try:
            # 首先清理
            if not self.clean():
                return False
            
            # 检查是否有uv
            if shutil.which("uv"):
                print("  使用UV构建...")
                result = subprocess.run(
                    ["uv", "build"],
                    cwd=self.project_root,
                    capture_output=True,
                    text=True
                )
            else:
                print("  使用传统工具构建...")
                # 确保安装了build工具
                subprocess.run([sys.executable, "-m", "pip", "install", "build"], check=True)
                result = subprocess.run(
                    [sys.executable, "-m", "build"],
                    cwd=self.project_root,
                    capture_output=True,
                    text=True
                )
            
            if result.returncode == 0:
                print("✅ 构建成功")
                self._list_build_artifacts()
                return True
            else:
                print(f"❌ 构建失败: {result.stderr}")
                return False
                
        except Exception as e:
            print(f"❌ 构建异常: {e}")
            return False
    
    def _list_build_artifacts(self) -> None:
        """列出构建产物"""
        if self.dist_dir.exists():
            print("\n📦 构建产物:")
            for file in self.dist_dir.iterdir():
                size = file.stat().st_size / 1024 / 1024  # MB
                print(f"  📄 {file.name} ({size:.2f} MB)")
    
    def test_installation(self) -> bool:
        """测试包安装
        
        Returns:
            bool: 测试是否成功
        """
        print("🧪 测试包安装...")
        
        if not self.dist_dir.exists() or not list(self.dist_dir.glob("*.whl")):
            print("❌ 未找到wheel文件，请先构建包")
            return False
        
        try:
            # 找到wheel文件
            wheel_file = next(self.dist_dir.glob("*.whl"))
            
            # 创建测试脚本
            test_script = self._create_test_script()
            
            # 在临时环境中测试
            print("  创建测试环境...")
            test_env = self.project_root / "test_env"
            
            # 创建虚拟环境
            subprocess.run([
                sys.executable, "-m", "venv", str(test_env)
            ], check=True)
            
            # 获取虚拟环境的Python路径
            if sys.platform == "win32":
                venv_python = test_env / "Scripts" / "python.exe"
                venv_pip = test_env / "Scripts" / "pip.exe"
            else:
                venv_python = test_env / "bin" / "python"
                venv_pip = test_env / "bin" / "pip"
            
            # 安装包
            print(f"  安装包: {wheel_file.name}")
            subprocess.run([
                str(venv_pip), "install", str(wheel_file)
            ], check=True)
            
            # 运行测试
            print("  运行功能测试...")
            result = subprocess.run([
                str(venv_python), "-c", test_script
            ], capture_output=True, text=True)
            
            # 清理测试环境
            shutil.rmtree(test_env)
            
            if result.returncode == 0:
                print("✅ 安装测试通过")
                print(f"  测试输出: {result.stdout.strip()}")
                return True
            else:
                print(f"❌ 安装测试失败: {result.stderr}")
                return False
                
        except Exception as e:
            print(f"❌ 测试异常: {e}")
            # 清理测试环境
            test_env = self.project_root / "test_env"
            if test_env.exists():
                shutil.rmtree(test_env)
            return False
    
    def _create_test_script(self) -> str:
        """创建测试脚本
        
        Returns:
            str: 测试脚本代码
        """
        return """
import sys
try:
    # 测试基本导入
    import boxmot
    print(f"✓ BoxMOT导入成功")
    
    # 测试StrongSort导入
    from boxmot.trackers.strongsort import StrongSort
    print(f"✓ StrongSort导入成功")
    
    # 测试创建追踪器
    from boxmot import create_tracker
    print(f"✓ create_tracker导入成功")
    
    print("🎉 所有测试通过")
    
except ImportError as e:
    print(f"❌ 导入失败: {e}")
    sys.exit(1)
except Exception as e:
    print(f"❌ 测试失败: {e}")
    sys.exit(1)
"""
    
    def check_package(self) -> bool:
        """检查包完整性
        
        Returns:
            bool: 检查是否通过
        """
        print("🔍 检查包完整性...")
        
        if not self.dist_dir.exists():
            print("❌ dist目录不存在")
            return False
        
        try:
            # 检查是否有twine
            if not shutil.which("twine"):
                print("  安装twine...")
                subprocess.run([sys.executable, "-m", "pip", "install", "twine"], check=True)
            
            # 使用twine检查
            result = subprocess.run([
                "twine", "check", "dist/*"
            ], cwd=self.project_root, capture_output=True, text=True)
            
            if result.returncode == 0:
                print("✅ 包检查通过")
                return True
            else:
                print(f"❌ 包检查失败: {result.stderr}")
                return False
                
        except Exception as e:
            print(f"❌ 检查异常: {e}")
            return False
    
    def publish(self, repository: str = "testpypi") -> bool:
        """发布包
        
        Args:
            repository: 发布仓库 ('testpypi' 或 'pypi')
            
        Returns:
            bool: 发布是否成功
        """
        print(f"📤 发布包到 {repository}...")
        
        # 检查包
        if not self.check_package():
            return False
        
        try:
            # 构建twine命令
            cmd = ["twine", "upload"]
            
            if repository == "testpypi":
                cmd.extend(["--repository", "testpypi"])
                print("  ⚠️  发布到测试环境")
            elif repository == "pypi":
                print("  🚨 发布到生产环境")
                # 确认发布
                confirm = input("  确认发布到生产环境？(yes/no): ")
                if confirm.lower() != "yes":
                    print("  取消发布")
                    return False
            else:
                print(f"❌ 未知仓库: {repository}")
                return False
            
            cmd.append("dist/*")
            
            # 执行发布
            result = subprocess.run(
                cmd,
                cwd=self.project_root,
                capture_output=True,
                text=True
            )
            
            if result.returncode == 0:
                print(f"✅ 发布成功到 {repository}")
                if repository == "testpypi":
                    print("  测试安装: pip install --index-url https://test.pypi.org/simple/ boxmot_with_tracker")
                else:
                    print("  安装命令: pip install boxmot_with_tracker")
                return True
            else:
                print(f"❌ 发布失败: {result.stderr}")
                return False
                
        except Exception as e:
            print(f"❌ 发布异常: {e}")
            return False


def main():
    """主函数"""
    parser = argparse.ArgumentParser(
        description="BoxMOT包构建脚本",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
示例:
  python build_package.py build          # 构建包
  python build_package.py test           # 测试安装
  python build_package.py check          # 检查包
  python build_package.py publish --test # 发布到测试环境
  python build_package.py publish --prod # 发布到生产环境
  python build_package.py all            # 执行完整流程
        """
    )
    
    parser.add_argument(
        "action",
        choices=["clean", "build", "test", "check", "publish", "all"],
        help="要执行的操作"
    )
    
    parser.add_argument(
        "--test",
        action="store_true",
        help="发布到测试环境 (TestPyPI)"
    )
    
    parser.add_argument(
        "--prod",
        action="store_true",
        help="发布到生产环境 (PyPI)"
    )
    
    args = parser.parse_args()
    
    # 获取项目根目录
    project_root = Path(__file__).parent
    builder = PackageBuilder(project_root)
    
    print(f"📁 项目目录: {project_root}")
    print(f"🎯 执行操作: {args.action}")
    print("=" * 50)
    
    success = True
    
    if args.action == "clean":
        success = builder.clean()
        
    elif args.action == "build":
        success = builder.build()
        
    elif args.action == "test":
        success = builder.test_installation()
        
    elif args.action == "check":
        success = builder.check_package()
        
    elif args.action == "publish":
        if args.test:
            success = builder.publish("testpypi")
        elif args.prod:
            success = builder.publish("pypi")
        else:
            print("❌ 请指定发布目标: --test 或 --prod")
            success = False
            
    elif args.action == "all":
        print("🚀 执行完整构建流程...\n")
        
        # 1. 构建
        success = builder.build()
        if not success:
            print("❌ 构建失败，停止流程")
            sys.exit(1)
        
        print("\n" + "=" * 50)
        
        # 2. 检查
        success = builder.check_package()
        if not success:
            print("❌ 包检查失败，停止流程")
            sys.exit(1)
        
        print("\n" + "=" * 50)
        
        # 3. 测试安装
        success = builder.test_installation()
        if not success:
            print("❌ 安装测试失败，停止流程")
            sys.exit(1)
        
        print("\n" + "=" * 50)
        print("🎉 完整流程执行成功！")
        print("\n📋 下一步:")
        print("  1. 发布到测试环境: python build_package.py publish --test")
        print("  2. 发布到生产环境: python build_package.py publish --prod")
    
    print("\n" + "=" * 50)
    if success:
        print("✅ 操作完成")
        sys.exit(0)
    else:
        print("❌ 操作失败")
        sys.exit(1)


if __name__ == "__main__":
    main()


""" run_burst.py

The example commands given below expect the following folder structure:

- data
    - gt
        - burst
            - {val,test}
                - all_classes
                    - all_classes.json (filename is irrelevant)
    - trackers
        - burst
            - exemplar_guided
                - {val,test}
                    - my_tracking_method
                        - data
                            - results.json  (filename is irrelevant)
            - class_guided
                - {val,test}
                    - my_other_tracking_method
                        - data
                            - results.json (filename is irrelevant)

Run example:

1) Exemplar-guided tasks (all three tasks share the same eval logic):
run_burst.py --USE_PARALLEL True --EXEMPLAR_GUIDED True --GT_FOLDER ../data/gt/burst/{val,test}/all_classes --TRACKERS_FOLDER ../data/trackers/burst/exemplar_guided/{val,test}

2) Class-guided tasks (common class and long-tail):
run_burst.py --USE_PARALLEL FTrue --EXEMPLAR_GUIDED False --GT_FOLDER ../data/gt/burst/{val,test}/all_classes --TRACKERS_FOLDER ../data/trackers/burst/class_guided/{val,test}

3) Refer to run_burst_ow.py for open world evaluation

Command Line Arguments: Defaults, # Comments
    Eval arguments:
        'USE_PARALLEL': False,
        'NUM_PARALLEL_CORES': 8,
        'BREAK_ON_ERROR': True,
        'PRINT_RESULTS': True,
        'PRINT_ONLY_COMBINED': False,
        'PRINT_CONFIG': True,
        'TIME_PROGRESS': True,
        'OUTPUT_SUMMARY': True,
        'OUTPUT_DETAILED': True,
        'PLOT_CURVES': True,
    Dataset arguments:
        'GT_FOLDER': os.path.join(code_path, 'data/gt/burst/val'),  # Location of GT data
        'TRACKERS_FOLDER': os.path.join(code_path, 'data/trackers/burst/class-guided/'),  # Trackers location
        'OUTPUT_FOLDER': None,  # Where to save eval results (if None, same as TRACKERS_FOLDER)
        'TRACKERS_TO_EVAL': None,  # Filenames of trackers to eval (if None, all in folder)
        'CLASSES_TO_EVAL': None,  # Classes to eval (if None, all classes)
        'SPLIT_TO_EVAL': 'training',  # Valid: 'training', 'val'
        'PRINT_CONFIG': True,  # Whether to print current config
        'TRACKER_SUB_FOLDER': 'data',  # Tracker files are in TRACKER_FOLDER/tracker_name/TRACKER_SUB_FOLDER
        'OUTPUT_SUB_FOLDER': '',  # Output files are saved in OUTPUT_FOLDER/tracker_name/OUTPUT_SUB_FOLDER
        'TRACKER_DISPLAY_NAMES': None,  # Names of trackers to display, if None: TRACKERS_TO_EVAL
        'MAX_DETECTIONS': 300,  # Number of maximal allowed detections per image (0 for unlimited)
    Metric arguments:
        'METRICS': ['HOTA', 'CLEAR', 'Identity', 'TrackMAP']
"""

import sys
import os
import argparse
from tabulate import tabulate
from multiprocessing import freeze_support

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import trackeval  # noqa: E402


def main():
    freeze_support()

    # Command line interface:
    default_eval_config = trackeval.Evaluator.get_default_eval_config()
    default_eval_config['PRINT_ONLY_COMBINED'] = True
    default_eval_config['DISPLAY_LESS_PROGRESS'] = True
    default_eval_config['PLOT_CURVES'] = False
    default_eval_config["OUTPUT_DETAILED"] = False
    default_eval_config["PRINT_RESULTS"] = False
    default_eval_config["OUTPUT_SUMMARY"] = False

    default_dataset_config = trackeval.datasets.BURST.get_default_dataset_config()

    # default_metrics_config = {'METRICS': ['HOTA', 'CLEAR', 'Identity', 'TrackMAP']}
    # default_metrics_config = {'METRICS': ['HOTA']}
    default_metrics_config = {'METRICS': ['HOTA', 'TrackMAP']}
    config = {**default_eval_config, **default_dataset_config, **default_metrics_config}  # Merge default configs
    parser = argparse.ArgumentParser()
    for setting in config.keys():
        if type(config[setting]) == list or type(config[setting]) == type(None):
            parser.add_argument("--" + setting, nargs='+')
        else:
            parser.add_argument("--" + setting)
    args = parser.parse_args().__dict__
    for setting in args.keys():
        if args[setting] is not None:
            if type(config[setting]) == type(True):
                if args[setting] == 'True':
                    x = True
                elif args[setting] == 'False':
                    x = False
                else:
                    raise Exception('Command line parameter ' + setting + 'must be True or False')
            elif type(config[setting]) == type(1):
                x = int(args[setting])
            elif type(args[setting]) == type(None):
                x = None
            else:
                x = args[setting]
            config[setting] = x
    eval_config = {k: v for k, v in config.items() if k in default_eval_config.keys()}
    dataset_config = {k: v for k, v in config.items() if k in default_dataset_config.keys()}
    metrics_config = {k: v for k, v in config.items() if k in default_metrics_config.keys()}

    # Run code
    evaluator = trackeval.Evaluator(eval_config)
    dataset_list = [trackeval.datasets.BURST(dataset_config)]
    metrics_list = []
    for metric in [trackeval.metrics.TrackMAP, trackeval.metrics.CLEAR, trackeval.metrics.Identity,
                   trackeval.metrics.HOTA]:
        if metric.get_name() in metrics_config['METRICS']:
            metrics_list.append(metric())
    if len(metrics_list) == 0:
        raise Exception('No metrics selected for evaluation')
    output_res, output_msg = evaluator.evaluate(dataset_list, metrics_list, show_progressbar=True)

    class_name_to_id = {x['name']: x['id'] for x in dataset_list[0].gt_data['categories']}
    known_list = [4, 13, 1038, 544, 1057, 34, 35, 36, 41, 45, 58, 60, 579, 1091, 1097, 1099, 78, 79, 81, 91, 1115,
                  1117, 95, 1122, 99, 1132, 621, 1135, 625, 118, 1144, 126, 642, 1155, 133, 1162, 139, 154, 174, 185,
                  699, 1215, 714, 717, 1229, 211, 729, 221, 229, 747, 235, 237, 779, 276, 805, 299, 829, 852, 347,
                  371, 382, 896, 392, 926, 937, 428, 429, 961, 452, 979, 980, 982, 475, 480, 993, 1001, 502, 1018]

    row_labels = ("HOTA", "DetA", "AssA", "AP")
    trackers = list(output_res['BURST'].keys())
    print("\n")

    def average_metric(m):
        return round(100*sum(m) / len(m), 2)

    for tracker in trackers:
        res = output_res['BURST'][tracker]['COMBINED_SEQ']
        all_names = [x for x in res.keys() if (x != 'cls_comb_cls_av') and (x != 'cls_comb_det_av')]

        class_split_names = {
            "All": [x for x in res.keys() if (x != 'cls_comb_cls_av') and (x != 'cls_comb_det_av')],
            "Common": [x for x in all_names if class_name_to_id[x] in known_list],
            "Uncommon": [x for x in all_names if class_name_to_id[x] not in known_list]
        }

        # table columns: 'all', 'common', 'uncommon'
        # table rows: HOTA, AssA, DetA, mAP
        table_data = []

        for row_label in row_labels:
            row = [row_label]
            for split_name in ["All", "Common", "Uncommon"]:
                split_classes = class_split_names[split_name]

                if row_label == "AP":
                    row.append(average_metric([res[c]['TrackMAP']["AP_all"].mean() for c in split_classes]))
                else:
                    row.append(average_metric([res[c]['HOTA'][row_label].mean() for c in split_classes]))

            table_data.append(row)

        print(f"Results for Tracker: {tracker}\n")
        print(tabulate(table_data, ["Metric", "All", "Common", "Uncommon"]))


if __name__ == '__main__':
    main()



from .hota import HOTA
from .clear import CLEAR
from .identity import Identity
from .count import Count
from .j_and_f import JAndF
from .track_map import TrackMAP
from .vace import VACE
from .ideucl import IDEucl


import os
import numpy as np
import json
import itertools
from collections import defaultdict
from scipy.optimize import linear_sum_assignment
from ..utils import TrackEvalException
from ._base_dataset import _BaseDataset
from .. import utils
from .. import _timing


class TAO_OW(_BaseDataset):
    """Dataset class for TAO tracking"""

    @staticmethod
    def get_default_dataset_config():
        """Default class config values"""
        code_path = utils.get_code_path()
        default_config = {
            'GT_FOLDER': os.path.join(code_path, 'data/gt/tao/tao_training'),  # Location of GT data
            'TRACKERS_FOLDER': os.path.join(code_path, 'data/trackers/tao/tao_training'),  # Trackers location
            'OUTPUT_FOLDER': None,  # Where to save eval results (if None, same as TRACKERS_FOLDER)
            'TRACKERS_TO_EVAL': None,  # Filenames of trackers to eval (if None, all in folder)
            'CLASSES_TO_EVAL': None,  # Classes to eval (if None, all classes)
            'SPLIT_TO_EVAL': 'training',  # Valid: 'training', 'val'
            'PRINT_CONFIG': True,  # Whether to print current config
            'TRACKER_SUB_FOLDER': 'data',  # Tracker files are in TRACKER_FOLDER/tracker_name/TRACKER_SUB_FOLDER
            'OUTPUT_SUB_FOLDER': '',  # Output files are saved in OUTPUT_FOLDER/tracker_name/OUTPUT_SUB_FOLDER
            'TRACKER_DISPLAY_NAMES': None,  # Names of trackers to display, if None: TRACKERS_TO_EVAL
            'MAX_DETECTIONS': 300,  # Number of maximal allowed detections per image (0 for unlimited)
            'SUBSET': 'all'
        }
        return default_config

    def __init__(self, config=None):
        """Initialise dataset, checking that all required files are present"""
        super().__init__()
        # Fill non-given config values with defaults
        self.config = utils.init_config(config, self.get_default_dataset_config(), self.get_name())
        self.gt_fol = self.config['GT_FOLDER']
        self.tracker_fol = self.config['TRACKERS_FOLDER']
        self.should_classes_combine = True
        self.use_super_categories = False

        self.tracker_sub_fol = self.config['TRACKER_SUB_FOLDER']
        self.output_fol = self.config['OUTPUT_FOLDER']
        if self.output_fol is None:
            self.output_fol = self.tracker_fol
        self.output_sub_fol = self.config['OUTPUT_SUB_FOLDER']

        gt_dir_files = [file for file in os.listdir(self.gt_fol) if file.endswith('.json')]
        if len(gt_dir_files) != 1:
            raise TrackEvalException(self.gt_fol + ' does not contain exactly one json file.')

        with open(os.path.join(self.gt_fol, gt_dir_files[0])) as f:
            self.gt_data = json.load(f)

        self.subset = self.config['SUBSET']
        if self.subset != 'all':
            # Split GT data into `known`, `unknown` or `distractor`
            self._split_known_unknown_distractor()
            self.gt_data = self._filter_gt_data(self.gt_data)

        # merge categories marked with a merged tag in TAO dataset
        self._merge_categories(self.gt_data['annotations'] + self.gt_data['tracks'])

        # Get sequences to eval and sequence information
        self.seq_list = [vid['name'].replace('/', '-') for vid in self.gt_data['videos']]
        self.seq_name_to_seq_id = {vid['name'].replace('/', '-'): vid['id'] for vid in self.gt_data['videos']}
        # compute mappings from videos to annotation data
        self.videos_to_gt_tracks, self.videos_to_gt_images = self._compute_vid_mappings(self.gt_data['annotations'])
        # compute sequence lengths
        self.seq_lengths = {vid['id']: 0 for vid in self.gt_data['videos']}
        for img in self.gt_data['images']:
            self.seq_lengths[img['video_id']] += 1
        self.seq_to_images_to_timestep = self._compute_image_to_timestep_mappings()
        self.seq_to_classes = {vid['id']: {'pos_cat_ids': list({track['category_id'] for track
                                                                in self.videos_to_gt_tracks[vid['id']]}),
                                           'neg_cat_ids': vid['neg_category_ids'],
                                           'not_exhaustively_labeled_cat_ids': vid['not_exhaustive_category_ids']}
                               for vid in self.gt_data['videos']}

        # Get classes to eval
        considered_vid_ids = [self.seq_name_to_seq_id[vid] for vid in self.seq_list]
        seen_cats = set([cat_id for vid_id in considered_vid_ids for cat_id
                         in self.seq_to_classes[vid_id]['pos_cat_ids']])
        # only classes with ground truth are evaluated in TAO
        self.valid_classes = [cls['name'] for cls in self.gt_data['categories'] if cls['id'] in seen_cats]
        # cls_name_to_cls_id_map = {cls['name']: cls['id'] for cls in self.gt_data['categories']}

        if self.config['CLASSES_TO_EVAL']:
            # self.class_list = [cls.lower() if cls.lower() in self.valid_classes else None
            #                    for cls in self.config['CLASSES_TO_EVAL']]
            self.class_list = ["object"]  # class-agnostic
            if not all(self.class_list):
                raise TrackEvalException('Attempted to evaluate an invalid class. Only classes ' +
                                         ', '.join(self.valid_classes) +
                                         ' are valid (classes present in ground truth data).')
        else:
            # self.class_list = [cls for cls in self.valid_classes]
            self.class_list = ["object"]  # class-agnostic
        # self.class_name_to_class_id = {k: v for k, v in cls_name_to_cls_id_map.items() if k in self.class_list}
        self.class_name_to_class_id = {"object": 1}  # class-agnostic

        # Get trackers to eval
        if self.config['TRACKERS_TO_EVAL'] is None:
            self.tracker_list = os.listdir(self.tracker_fol)
        else:
            self.tracker_list = self.config['TRACKERS_TO_EVAL']

        if self.config['TRACKER_DISPLAY_NAMES'] is None:
            self.tracker_to_disp = dict(zip(self.tracker_list, self.tracker_list))
        elif (self.config['TRACKERS_TO_EVAL'] is not None) and (
                len(self.config['TRACKER_DISPLAY_NAMES']) == len(self.tracker_list)):
            self.tracker_to_disp = dict(zip(self.tracker_list, self.config['TRACKER_DISPLAY_NAMES']))
        else:
            raise TrackEvalException('List of tracker files and tracker display names do not match.')

        self.tracker_data = {tracker: dict() for tracker in self.tracker_list}

        for tracker in self.tracker_list:
            tr_dir_files = [file for file in os.listdir(os.path.join(self.tracker_fol, tracker, self.tracker_sub_fol))
                            if file.endswith('.json')]
            if len(tr_dir_files) != 1:
                raise TrackEvalException(os.path.join(self.tracker_fol, tracker, self.tracker_sub_fol)
                                         + ' does not contain exactly one json file.')
            with open(os.path.join(self.tracker_fol, tracker, self.tracker_sub_fol, tr_dir_files[0])) as f:
                curr_data = json.load(f)

            # limit detections if MAX_DETECTIONS > 0
            if self.config['MAX_DETECTIONS']:
                curr_data = self._limit_dets_per_image(curr_data)

            # fill missing video ids
            self._fill_video_ids_inplace(curr_data)

            # make track ids unique over whole evaluation set
            self._make_track_ids_unique(curr_data)

            # merge categories marked with a merged tag in TAO dataset
            self._merge_categories(curr_data)

            # get tracker sequence information
            curr_videos_to_tracker_tracks, curr_videos_to_tracker_images = self._compute_vid_mappings(curr_data)
            self.tracker_data[tracker]['vids_to_tracks'] = curr_videos_to_tracker_tracks
            self.tracker_data[tracker]['vids_to_images'] = curr_videos_to_tracker_images

    def get_display_name(self, tracker):
        return self.tracker_to_disp[tracker]

    def _load_raw_file(self, tracker, seq, is_gt):
        """Load a file (gt or tracker) in the TAO format

        If is_gt, this returns a dict which contains the fields:
        [gt_ids, gt_classes] : list (for each timestep) of 1D NDArrays (for each det).
        [gt_dets]: list (for each timestep) of lists of detections.
        [classes_to_gt_tracks]: dictionary with class values as keys and list of dictionaries (with frame indices as
                                keys and corresponding segmentations as values) for each track
        [classes_to_gt_track_ids, classes_to_gt_track_areas, classes_to_gt_track_lengths]: dictionary with class values
                                as keys and lists (for each track) as values

        if not is_gt, this returns a dict which contains the fields:
        [tracker_ids, tracker_classes, tracker_confidences] : list (for each timestep) of 1D NDArrays (for each det).
        [tracker_dets]: list (for each timestep) of lists of detections.
        [classes_to_dt_tracks]: dictionary with class values as keys and list of dictionaries (with frame indices as
                                keys and corresponding segmentations as values) for each track
        [classes_to_dt_track_ids, classes_to_dt_track_areas, classes_to_dt_track_lengths]: dictionary with class values
                                                                                           as keys and lists as values
        [classes_to_dt_track_scores]: dictionary with class values as keys and 1D numpy arrays as values
        """
        seq_id = self.seq_name_to_seq_id[seq]
        # File location
        if is_gt:
            imgs = self.videos_to_gt_images[seq_id]
        else:
            imgs = self.tracker_data[tracker]['vids_to_images'][seq_id]

        # Convert data to required format
        num_timesteps = self.seq_lengths[seq_id]
        img_to_timestep = self.seq_to_images_to_timestep[seq_id]
        data_keys = ['ids', 'classes', 'dets']
        if not is_gt:
            data_keys += ['tracker_confidences']
        raw_data = {key: [None] * num_timesteps for key in data_keys}
        for img in imgs:
            # some tracker data contains images without any ground truth information, these are ignored
            try:
                t = img_to_timestep[img['id']]
            except KeyError:
                continue
            annotations = img['annotations']
            raw_data['dets'][t] = np.atleast_2d([ann['bbox'] for ann in annotations]).astype(float)
            raw_data['ids'][t] = np.atleast_1d([ann['track_id'] for ann in annotations]).astype(int)
            raw_data['classes'][t] = np.atleast_1d([1 for _ in annotations]).astype(int)   # class-agnostic
            if not is_gt:
                raw_data['tracker_confidences'][t] = np.atleast_1d([ann['score'] for ann in annotations]).astype(float)

        for t, d in enumerate(raw_data['dets']):
            if d is None:
                raw_data['dets'][t] = np.empty((0, 4)).astype(float)
                raw_data['ids'][t] = np.empty(0).astype(int)
                raw_data['classes'][t] = np.empty(0).astype(int)
                if not is_gt:
                    raw_data['tracker_confidences'][t] = np.empty(0)

        if is_gt:
            key_map = {'ids': 'gt_ids',
                       'classes': 'gt_classes',
                       'dets': 'gt_dets'}
        else:
            key_map = {'ids': 'tracker_ids',
                       'classes': 'tracker_classes',
                       'dets': 'tracker_dets'}
        for k, v in key_map.items():
            raw_data[v] = raw_data.pop(k)

        # all_classes = [self.class_name_to_class_id[cls] for cls in self.class_list]
        all_classes = [1]  # class-agnostic

        if is_gt:
            classes_to_consider = all_classes
            all_tracks = self.videos_to_gt_tracks[seq_id]
        else:
            # classes_to_consider = self.seq_to_classes[seq_id]['pos_cat_ids'] \
            #                       + self.seq_to_classes[seq_id]['neg_cat_ids']
            classes_to_consider = all_classes  # class-agnostic
            all_tracks = self.tracker_data[tracker]['vids_to_tracks'][seq_id]

        # classes_to_tracks = {cls: [track for track in all_tracks if track['category_id'] == cls]
        #                      if cls in classes_to_consider else [] for cls in all_classes}
        classes_to_tracks = {cls: [track for track in all_tracks]
        if cls in classes_to_consider else [] for cls in all_classes}  # class-agnostic

        # mapping from classes to track information
        raw_data['classes_to_tracks'] = {cls: [{det['image_id']: np.atleast_1d(det['bbox'])
                                                for det in track['annotations']} for track in tracks]
                                         for cls, tracks in classes_to_tracks.items()}
        raw_data['classes_to_track_ids'] = {cls: [track['id'] for track in tracks]
                                            for cls, tracks in classes_to_tracks.items()}
        raw_data['classes_to_track_areas'] = {cls: [track['area'] for track in tracks]
                                              for cls, tracks in classes_to_tracks.items()}
        raw_data['classes_to_track_lengths'] = {cls: [len(track['annotations']) for track in tracks]
                                                for cls, tracks in classes_to_tracks.items()}

        if not is_gt:
            raw_data['classes_to_dt_track_scores'] = {cls: np.array([np.mean([float(x['score'])
                                                                              for x in track['annotations']])
                                                                     for track in tracks])
                                                      for cls, tracks in classes_to_tracks.items()}

        if is_gt:
            key_map = {'classes_to_tracks': 'classes_to_gt_tracks',
                       'classes_to_track_ids': 'classes_to_gt_track_ids',
                       'classes_to_track_lengths': 'classes_to_gt_track_lengths',
                       'classes_to_track_areas': 'classes_to_gt_track_areas'}
        else:
            key_map = {'classes_to_tracks': 'classes_to_dt_tracks',
                       'classes_to_track_ids': 'classes_to_dt_track_ids',
                       'classes_to_track_lengths': 'classes_to_dt_track_lengths',
                       'classes_to_track_areas': 'classes_to_dt_track_areas'}
        for k, v in key_map.items():
            raw_data[v] = raw_data.pop(k)

        raw_data['num_timesteps'] = num_timesteps
        raw_data['neg_cat_ids'] = self.seq_to_classes[seq_id]['neg_cat_ids']
        raw_data['not_exhaustively_labeled_cls'] = self.seq_to_classes[seq_id]['not_exhaustively_labeled_cat_ids']
        raw_data['seq'] = seq
        return raw_data

    @_timing.time
    def get_preprocessed_seq_data(self, raw_data, cls):
        """ Preprocess data for a single sequence for a single class ready for evaluation.
        Inputs:
             - raw_data is a dict containing the data for the sequence already read in by get_raw_seq_data().
             - cls is the class to be evaluated.
        Outputs:
             - data is a dict containing all of the information that metrics need to perform evaluation.
                It contains the following fields:
                    [num_timesteps, num_gt_ids, num_tracker_ids, num_gt_dets, num_tracker_dets] : integers.
                    [gt_ids, tracker_ids, tracker_confidences]: list (for each timestep) of 1D NDArrays (for each det).
                    [gt_dets, tracker_dets]: list (for each timestep) of lists of detections.
                    [similarity_scores]: list (for each timestep) of 2D NDArrays.
        Notes:
            General preprocessing (preproc) occurs in 4 steps. Some datasets may not use all of these steps.
                1) Extract only detections relevant for the class to be evaluated (including distractor detections).
                2) Match gt dets and tracker dets. Remove tracker dets that are matched to a gt det that is of a
                    distractor class, or otherwise marked as to be removed.
                3) Remove unmatched tracker dets if they fall within a crowd ignore region or don't meet a certain
                    other criteria (e.g. are too small).
                4) Remove gt dets that were only useful for preprocessing and not for actual evaluation.
            After the above preprocessing steps, this function also calculates the number of gt and tracker detections
                and unique track ids. It also relabels gt and tracker ids to be contiguous and checks that ids are
                unique within each timestep.
        TAO:
            In TAO, the 4 preproc steps are as follow:
                1) All classes present in the ground truth data are evaluated separately.
                2) No matched tracker detections are removed.
                3) Unmatched tracker detections are removed if there is not ground truth data and the class does not
                    belong to the categories marked as negative for this sequence. Additionally, unmatched tracker
                    detections for classes which are marked as not exhaustively labeled are removed.
                4) No gt detections are removed.
            Further, for TrackMAP computation track representations for the given class are accessed from a dictionary
            and the tracks from the tracker data are sorted according to the tracker confidence.
        """
        cls_id = self.class_name_to_class_id[cls]
        is_not_exhaustively_labeled = cls_id in raw_data['not_exhaustively_labeled_cls']
        is_neg_category = cls_id in raw_data['neg_cat_ids']

        data_keys = ['gt_ids', 'tracker_ids', 'gt_dets', 'tracker_dets', 'tracker_confidences', 'similarity_scores']
        data = {key: [None] * raw_data['num_timesteps'] for key in data_keys}
        unique_gt_ids = []
        unique_tracker_ids = []
        num_gt_dets = 0
        num_tracker_dets = 0
        for t in range(raw_data['num_timesteps']):

            # Only extract relevant dets for this class for preproc and eval (cls)
            gt_class_mask = np.atleast_1d(raw_data['gt_classes'][t] == cls_id)
            gt_class_mask = gt_class_mask.astype(bool)
            gt_ids = raw_data['gt_ids'][t][gt_class_mask]
            gt_dets = raw_data['gt_dets'][t][gt_class_mask]

            tracker_class_mask = np.atleast_1d(raw_data['tracker_classes'][t] == cls_id)
            tracker_class_mask = tracker_class_mask.astype(bool)
            tracker_ids = raw_data['tracker_ids'][t][tracker_class_mask]
            tracker_dets = raw_data['tracker_dets'][t][tracker_class_mask]
            tracker_confidences = raw_data['tracker_confidences'][t][tracker_class_mask]
            similarity_scores = raw_data['similarity_scores'][t][gt_class_mask, :][:, tracker_class_mask]

            # Match tracker and gt dets (with hungarian algorithm).
            unmatched_indices = np.arange(tracker_ids.shape[0])
            if gt_ids.shape[0] > 0 and tracker_ids.shape[0] > 0:
                matching_scores = similarity_scores.copy()
                matching_scores[matching_scores < 0.5 - np.finfo('float').eps] = 0
                match_rows, match_cols = linear_sum_assignment(-matching_scores)
                actually_matched_mask = matching_scores[match_rows, match_cols] > 0 + np.finfo('float').eps
                match_cols = match_cols[actually_matched_mask]
                unmatched_indices = np.delete(unmatched_indices, match_cols, axis=0)

            if gt_ids.shape[0] == 0 and not is_neg_category:
                to_remove_tracker = unmatched_indices
            elif is_not_exhaustively_labeled:
                to_remove_tracker = unmatched_indices
            else:
                to_remove_tracker = np.array([], dtype=int)

            # remove all unwanted unmatched tracker detections
            data['tracker_ids'][t] = np.delete(tracker_ids, to_remove_tracker, axis=0)
            data['tracker_dets'][t] = np.delete(tracker_dets, to_remove_tracker, axis=0)
            data['tracker_confidences'][t] = np.delete(tracker_confidences, to_remove_tracker, axis=0)
            similarity_scores = np.delete(similarity_scores, to_remove_tracker, axis=1)

            data['gt_ids'][t] = gt_ids
            data['gt_dets'][t] = gt_dets
            data['similarity_scores'][t] = similarity_scores

            unique_gt_ids += list(np.unique(data['gt_ids'][t]))
            unique_tracker_ids += list(np.unique(data['tracker_ids'][t]))
            num_tracker_dets += len(data['tracker_ids'][t])
            num_gt_dets += len(data['gt_ids'][t])

        # Re-label IDs such that there are no empty IDs
        if len(unique_gt_ids) > 0:
            unique_gt_ids = np.unique(unique_gt_ids)
            gt_id_map = np.nan * np.ones((np.max(unique_gt_ids) + 1))
            gt_id_map[unique_gt_ids] = np.arange(len(unique_gt_ids))
            for t in range(raw_data['num_timesteps']):
                if len(data['gt_ids'][t]) > 0:
                    data['gt_ids'][t] = gt_id_map[data['gt_ids'][t]].astype(int)
        if len(unique_tracker_ids) > 0:
            unique_tracker_ids = np.unique(unique_tracker_ids)
            tracker_id_map = np.nan * np.ones((np.max(unique_tracker_ids) + 1))
            tracker_id_map[unique_tracker_ids] = np.arange(len(unique_tracker_ids))
            for t in range(raw_data['num_timesteps']):
                if len(data['tracker_ids'][t]) > 0:
                    data['tracker_ids'][t] = tracker_id_map[data['tracker_ids'][t]].astype(int)

        # Record overview statistics.
        data['num_tracker_dets'] = num_tracker_dets
        data['num_gt_dets'] = num_gt_dets
        data['num_tracker_ids'] = len(unique_tracker_ids)
        data['num_gt_ids'] = len(unique_gt_ids)
        data['num_timesteps'] = raw_data['num_timesteps']
        data['seq'] = raw_data['seq']

        # get track representations
        data['gt_tracks'] = raw_data['classes_to_gt_tracks'][cls_id]
        data['gt_track_ids'] = raw_data['classes_to_gt_track_ids'][cls_id]
        data['gt_track_lengths'] = raw_data['classes_to_gt_track_lengths'][cls_id]
        data['gt_track_areas'] = raw_data['classes_to_gt_track_areas'][cls_id]
        data['dt_tracks'] = raw_data['classes_to_dt_tracks'][cls_id]
        data['dt_track_ids'] = raw_data['classes_to_dt_track_ids'][cls_id]
        data['dt_track_lengths'] = raw_data['classes_to_dt_track_lengths'][cls_id]
        data['dt_track_areas'] = raw_data['classes_to_dt_track_areas'][cls_id]
        data['dt_track_scores'] = raw_data['classes_to_dt_track_scores'][cls_id]
        data['not_exhaustively_labeled'] = is_not_exhaustively_labeled
        data['iou_type'] = 'bbox'

        # sort tracker data tracks by tracker confidence scores
        if data['dt_tracks']:
            idx = np.argsort([-score for score in data['dt_track_scores']], kind="mergesort")
            data['dt_track_scores'] = [data['dt_track_scores'][i] for i in idx]
            data['dt_tracks'] = [data['dt_tracks'][i] for i in idx]
            data['dt_track_ids'] = [data['dt_track_ids'][i] for i in idx]
            data['dt_track_lengths'] = [data['dt_track_lengths'][i] for i in idx]
            data['dt_track_areas'] = [data['dt_track_areas'][i] for i in idx]
        # Ensure that ids are unique per timestep.
        self._check_unique_ids(data)

        return data

    def _calculate_similarities(self, gt_dets_t, tracker_dets_t):
        similarity_scores = self._calculate_box_ious(gt_dets_t, tracker_dets_t)
        return similarity_scores

    def _merge_categories(self, annotations):
        """
        Merges categories with a merged tag. Adapted from https://github.com/TAO-Dataset
        :param annotations: the annotations in which the classes should be merged
        :return: None
        """
        merge_map = {}
        for category in self.gt_data['categories']:
            if 'merged' in category:
                for to_merge in category['merged']:
                    merge_map[to_merge['id']] = category['id']

        for ann in annotations:
            ann['category_id'] = merge_map.get(ann['category_id'], ann['category_id'])

    def _compute_vid_mappings(self, annotations):
        """
        Computes mappings from Videos to corresponding tracks and images.
        :param annotations: the annotations for which the mapping should be generated
        :return: the video-to-track-mapping, the video-to-image-mapping
        """
        vids_to_tracks = {}
        vids_to_imgs = {}
        vid_ids = [vid['id'] for vid in self.gt_data['videos']]

        # compute an mapping from image IDs to images
        images = {}
        for image in self.gt_data['images']:
            images[image['id']] = image

        for ann in annotations:
            ann["area"] = ann["bbox"][2] * ann["bbox"][3]

            vid = ann["video_id"]
            if ann["video_id"] not in vids_to_tracks.keys():
                vids_to_tracks[ann["video_id"]] = list()
            if ann["video_id"] not in vids_to_imgs.keys():
                vids_to_imgs[ann["video_id"]] = list()

            # Fill in vids_to_tracks
            tid = ann["track_id"]
            exist_tids = [track["id"] for track in vids_to_tracks[vid]]
            try:
                index1 = exist_tids.index(tid)
            except ValueError:
                index1 = -1
            if tid not in exist_tids:
                curr_track = {"id": tid, "category_id": ann['category_id'],
                              "video_id": vid, "annotations": [ann]}
                vids_to_tracks[vid].append(curr_track)
            else:
                vids_to_tracks[vid][index1]["annotations"].append(ann)

            # Fill in vids_to_imgs
            img_id = ann['image_id']
            exist_img_ids = [img["id"] for img in vids_to_imgs[vid]]
            try:
                index2 = exist_img_ids.index(img_id)
            except ValueError:
                index2 = -1
            if index2 == -1:
                curr_img = {"id": img_id, "annotations": [ann]}
                vids_to_imgs[vid].append(curr_img)
            else:
                vids_to_imgs[vid][index2]["annotations"].append(ann)

        # sort annotations by frame index and compute track area
        for vid, tracks in vids_to_tracks.items():
            for track in tracks:
                track["annotations"] = sorted(
                    track['annotations'],
                    key=lambda x: images[x['image_id']]['frame_index'])
                # Computer average area
                track["area"] = (sum(x['area'] for x in track['annotations']) / len(track['annotations']))

        # Ensure all videos are present
        for vid_id in vid_ids:
            if vid_id not in vids_to_tracks.keys():
                vids_to_tracks[vid_id] = []
            if vid_id not in vids_to_imgs.keys():
                vids_to_imgs[vid_id] = []

        return vids_to_tracks, vids_to_imgs

    def _compute_image_to_timestep_mappings(self):
        """
        Computes a mapping from images to the corresponding timestep in the sequence.
        :return: the image-to-timestep-mapping
        """
        images = {}
        for image in self.gt_data['images']:
            images[image['id']] = image

        seq_to_imgs_to_timestep = {vid['id']: dict() for vid in self.gt_data['videos']}
        for vid in seq_to_imgs_to_timestep:
            curr_imgs = [img['id'] for img in self.videos_to_gt_images[vid]]
            curr_imgs = sorted(curr_imgs, key=lambda x: images[x]['frame_index'])
            seq_to_imgs_to_timestep[vid] = {curr_imgs[i]: i for i in range(len(curr_imgs))}

        return seq_to_imgs_to_timestep

    def _limit_dets_per_image(self, annotations):
        """
        Limits the number of detections for each image to config['MAX_DETECTIONS']. Adapted from
        https://github.com/TAO-Dataset/
        :param annotations: the annotations in which the detections should be limited
        :return: the annotations with limited detections
        """
        max_dets = self.config['MAX_DETECTIONS']
        img_ann = defaultdict(list)
        for ann in annotations:
            img_ann[ann["image_id"]].append(ann)

        for img_id, _anns in img_ann.items():
            if len(_anns) <= max_dets:
                continue
            _anns = sorted(_anns, key=lambda x: x["score"], reverse=True)
            img_ann[img_id] = _anns[:max_dets]

        return [ann for anns in img_ann.values() for ann in anns]

    def _fill_video_ids_inplace(self, annotations):
        """
        Fills in missing video IDs inplace. Adapted from https://github.com/TAO-Dataset/
        :param annotations: the annotations for which the videos IDs should be filled inplace
        :return: None
        """
        missing_video_id = [x for x in annotations if 'video_id' not in x]
        if missing_video_id:
            image_id_to_video_id = {
                x['id']: x['video_id'] for x in self.gt_data['images']
            }
            for x in missing_video_id:
                x['video_id'] = image_id_to_video_id[x['image_id']]

    @staticmethod
    def _make_track_ids_unique(annotations):
        """
        Makes the track IDs unqiue over the whole annotation set. Adapted from https://github.com/TAO-Dataset/
        :param annotations: the annotation set
        :return: the number of updated IDs
        """
        track_id_videos = {}
        track_ids_to_update = set()
        max_track_id = 0
        for ann in annotations:
            t = ann['track_id']
            if t not in track_id_videos:
                track_id_videos[t] = ann['video_id']

            if ann['video_id'] != track_id_videos[t]:
                # Track id is assigned to multiple videos
                track_ids_to_update.add(t)
            max_track_id = max(max_track_id, t)

        if track_ids_to_update:
            print('true')
            next_id = itertools.count(max_track_id + 1)
            new_track_ids = defaultdict(lambda: next(next_id))
            for ann in annotations:
                t = ann['track_id']
                v = ann['video_id']
                if t in track_ids_to_update:
                    ann['track_id'] = new_track_ids[t, v]
        return len(track_ids_to_update)

    def _split_known_unknown_distractor(self):
        all_ids = set([i for i in range(1, 2000)])  # 2000 is larger than the max category id in TAO-OW.
        # `knowns` includes 78 TAO_category_ids that corresponds to 78 COCO classes.
        # (The other 2 COCO classes do not have corresponding classes in TAO).
        self.knowns = {4, 13, 1038, 544, 1057, 34, 35, 36, 41, 45, 58, 60, 579, 1091, 1097, 1099, 78, 79, 81, 91, 1115,
                     1117, 95, 1122, 99, 1132, 621, 1135, 625, 118, 1144, 126, 642, 1155, 133, 1162, 139, 154, 174, 185,
                     699, 1215, 714, 717, 1229, 211, 729, 221, 229, 747, 235, 237, 779, 276, 805, 299, 829, 852, 347,
                     371, 382, 896, 392, 926, 937, 428, 429, 961, 452, 979, 980, 982, 475, 480, 993, 1001, 502, 1018}
        # `distractors` is defined as in the paper "Opening up Open-World Tracking"
        self.distractors = {20, 63, 108, 180, 188, 204, 212, 247, 303, 403, 407, 415, 490, 504, 507, 513, 529, 567,
                            569, 588, 672, 691, 702, 708, 711, 720, 736, 737, 798, 813, 815, 827, 831, 851, 877, 883,
                            912, 971, 976, 1130, 1133, 1134, 1169, 1184, 1220}
        self.unknowns = all_ids.difference(self.knowns.union(self.distractors))

    def _filter_gt_data(self, raw_gt_data):
        """
        Filter out irrelevant data in the raw_gt_data
        Args:
            raw_gt_data: directly loaded from json.

        Returns:
            filtered gt_data
        """
        valid_cat_ids = list()
        if self.subset == "known":
            valid_cat_ids = self.knowns
        elif self.subset == "distractor":
            valid_cat_ids = self.distractors
        elif self.subset == "unknown":
            valid_cat_ids = self.unknowns
        # elif self.subset == "test_only_unknowns":
        #     valid_cat_ids = test_only_unknowns
        else:
            raise Exception("The parameter `SUBSET` is incorrect")

        filtered = dict()
        filtered["videos"] = raw_gt_data["videos"]
        # filtered["videos"] = list()
        unwanted_vid = set()
        # for video in raw_gt_data["videos"]:
        #     datasrc = video["name"].split('/')[1]
        #     if datasrc in data_srcs:
        #         filtered["videos"].append(video)
        #     else:
        #         unwanted_vid.add(video["id"])

        filtered["annotations"] = list()
        for ann in raw_gt_data["annotations"]:
            if (ann["video_id"] not in unwanted_vid) and (ann["category_id"] in valid_cat_ids):
                filtered["annotations"].append(ann)

        filtered["tracks"] = list()
        for track in raw_gt_data["tracks"]:
            if (track["video_id"] not in unwanted_vid) and (track["category_id"] in valid_cat_ids):
                filtered["tracks"].append(track)

        filtered["images"] = list()
        for image in raw_gt_data["images"]:
            if image["video_id"] not in unwanted_vid:
                filtered["images"].append(image)

        filtered["categories"] = list()
        for cat in raw_gt_data["categories"]:
            if cat["id"] in valid_cat_ids:
                filtered["categories"].append(cat)

        filtered["info"] = raw_gt_data["info"]
        filtered["licenses"] = raw_gt_data["licenses"]

        return filtered



#!/usr/bin/env python3
import cv2
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from pathlib import Path
from typing import Union
import argparse

def plot_gt_boxes_with_trajectories(
    seq_dir: Union[str, Path],
    use_temp_gt: bool = True,
    pad: int = 0
):
    """
    Plot all ground-truth boxes for a MOT17 sequence, coloring each object ID
    with a unique color and drawing its trajectory through the centers of its boxes.

    Args:
        seq_dir (str | Path): Path to the sequence folder (containing img1/ and gt/).
        use_temp_gt (bool): If True, plot gt/gt_temp.txt (filtered by FPS).
                            Otherwise plot gt/gt.txt.
        pad (int): Extra padding (in pixels) to add around the outermost boxes.
    """
    seq_dir = Path(seq_dir)
    # --- 1) grab image size from first frame ---
    img_files = sorted((seq_dir / "img1").glob("*.jpg"))
    if not img_files:
        raise RuntimeError(f"No images found in {seq_dir / 'img1'}")
    first_img = img_files[0]
    img = cv2.imread(str(first_img))
    if img is None:
        raise RuntimeError(f"Could not load image {first_img}")
    height, width = img.shape[:2]

    # --- 2) load the GT file ---
    gt_file = seq_dir / "gt" / ("gt_temp.txt" if use_temp_gt else "gt.txt")
    orig_gt = np.loadtxt(gt_file, delimiter=",")
    MOT_DISTRACTOR_IDS = [2, 7, 8, 12, 13]  # e.g., person_on_vehicle, static_person, etc.
    class_ids = orig_gt[:, 1].astype(int)
    mask = ~np.isin(class_ids, MOT_DISTRACTOR_IDS)
    orig_gt = orig_gt[mask]
    # we'll unpack the first 6 columns for plotting

    # --- 3) compute the full extents of all boxes ---
    boxes = orig_gt[:, 2:6]  # x, y, w, h
    xs      = boxes[:, 0]
    ys      = boxes[:, 1]
    rights  = xs + boxes[:, 2]
    bottoms = ys + boxes[:, 3]

    xmin = min(xs.min(), 0) - pad
    ymin = min(ys.min(), 0) - pad
    xmax = max(rights.max(), width) + pad
    ymax = max(bottoms.max(), height) + pad

    # --- 4) set up Matplotlib figure ---
    fig, ax = plt.subplots(
        figsize=(10, 10 * (ymax - ymin) / (xmax - xmin))
    )
    ax.set_xlim(xmin, xmax)
    # invert y so origin is top-left
    ax.set_ylim(ymax, ymin)
    ax.set_aspect('equal')
    ax.set_title(
        f"GT boxes & trajectories for sequence {seq_dir.name}\n"
        f"Image size: {width}×{height}, total boxes: {len(boxes)}"
    )
    ax.set_xlabel("x")
    ax.set_ylabel("y")

    # --- 5) draw the original image and its border ---
    ax.imshow(
        cv2.cvtColor(img, cv2.COLOR_BGR2RGB),
        extent=(0, width, height, 0),
        zorder=0
    )
    img_border = patches.Rectangle(
        (0, 0), width, height,
        linewidth=1, edgecolor='gray', linestyle='--',
        facecolor='none', zorder=1
    )
    ax.add_patch(img_border)

    # --- 6) plot each box in a unique color per ID ---
    ids = orig_gt[:, 1].astype(int)
    unique_ids = np.unique(ids)
    cmap = plt.colormaps.get_cmap('tab20')
    colors = [cmap(i / len(unique_ids)) for i in range(len(unique_ids))]
    id2color = {obj_id: colors[i] for i, obj_id in enumerate(unique_ids)}

    for row in orig_gt:
        frame, obj_id, x, y, w_box, h_box = row[:6]
        obj_id = int(obj_id)
        color = id2color[obj_id]
        outside = (
            x < 0 or y < 0 or
            x + w_box > width or
            y + h_box > height
        )
        rect = patches.Rectangle(
            (x, y),
            w_box, h_box,
            linewidth=1,
            edgecolor=color,
            facecolor='none',
            linestyle='--' if outside else '-',
            zorder=2
        )
        ax.add_patch(rect)

    # annotate count of out‐of‐frame boxes
    n_outside = sum(
        (row[2] < 0 or row[3] < 0 or
         row[2] + row[4] > width or
         row[3] + row[5] > height)
        for row in orig_gt
    )
    ax.text(
        xmin + 0.01*(xmax - xmin),
        ymax - 0.02*(ymax - ymin),
        f"Outside boxes: {n_outside} / {len(boxes)}",
        color='red', fontsize=12, va='top', zorder=3
    )

    # --- 7) plot each ID’s trajectory via box centers ---
    for obj_id in unique_ids:
        sel = orig_gt[orig_gt[:, 1] == obj_id]
        # center = (x + w/2, y + h/2)
        centers = np.stack([
            sel[:, 2] + sel[:, 4] / 2,
            sel[:, 3] + sel[:, 5] / 2
        ], axis=1)
        # sort by frame index
        order = np.argsort(sel[:, 0].astype(int))
        centers = centers[order]
        ax.plot(
            centers[:, 0], centers[:, 1],
            '-', linewidth=1,
            color=id2color[obj_id],
            zorder=3
        )

    # (optional) legend if number of IDs is small
    # ax.legend(
    #     [f"ID {i}" for i in unique_ids],
    #     loc='upper right',
    #     fontsize='small',
    #     ncol=2,
    #     framealpha=0.5
    # )

    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Plot MOT17 ground-truth boxes and trajectories"
    )
    parser.add_argument(
        "seq_dir",
        help="Path to the MOT17 sequence folder (must contain img1/ and gt/ subfolders)"
    )
    parser.add_argument(
        "--use_temp_gt",
        action="store_true",
        help="Plot gt/gt_temp.txt instead of gt/gt.txt"
    )
    parser.add_argument(
        "--pad",
        type=int,
        default=0,
        help="Extra padding (pixels) around the outermost boxes"
    )
    args = parser.parse_args()

    plot_gt_boxes_with_trajectories(
        args.seq_dir,
        use_temp_gt=args.use_temp_gt,
        pad=args.pad
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
专门测试 StrongSort._remove_duplicate_tracks 方法的单元测试

该测试文件专注于验证重复轨迹ID处理逻辑的正确性，包括：
- 无重复ID的情况
- 存在重复ID的情况
- 相同time_since_update值的处理
- 空轨迹列表的处理
- 多个重复ID组的处理
"""

import unittest
import numpy as np
from unittest.mock import Mock, patch
from pathlib import Path
import torch

# 导入被测试的类
from boxmot.trackers.strongsort.strongsort import StrongSort


class TestRemoveDuplicateTracks(unittest.TestCase):
    """StrongSort._remove_duplicate_tracks方法的专项单元测试"""

    def setUp(self):
        """测试前的设置"""
        # Mock所有外部依赖
        with patch('boxmot.trackers.strongsort.strongsort.ReidAutoBackend') as mock_reid, \
             patch('boxmot.trackers.strongsort.strongsort.Tracker') as mock_tracker_class, \
             patch('boxmot.trackers.strongsort.strongsort.get_cmc_method') as mock_cmc_method:
            
            # 配置mock对象
            mock_model = Mock()
            mock_model.get_features.return_value = np.random.rand(0, 128)
            mock_reid.return_value.model = mock_model
            
            mock_tracker = Mock()
            mock_tracker.tracks = []
            mock_tracker_class.return_value = mock_tracker
            
            mock_cmc = Mock()
            mock_cmc.apply.return_value = np.eye(3)
            mock_cmc_method.return_value = lambda: mock_cmc
            
            # 创建StrongSort实例
            self.strongsort = StrongSort(
                reid_weights=Path("dummy.pt"),
                device=torch.device("cpu"),
                half=False,
                per_class=False,
                min_conf=0.5
            )
            
            # 保存mock对象的引用
            self.mock_tracker = self.strongsort.tracker

    def test_remove_duplicate_tracks_no_duplicates(self):
        """测试没有重复ID的情况"""
        # 创建不同ID的轨迹
        mock_tracks = []
        for i in range(3):
            mock_track = Mock()
            mock_track.id = i + 1
            mock_track.time_since_update = i
            mock_tracks.append(mock_track)
        
        self.mock_tracker.tracks = mock_tracks
        original_count = len(self.mock_tracker.tracks)
        
        # 调用方法
        self.strongsort._remove_duplicate_tracks()
        
        # 验证没有轨迹被移除
        self.assertEqual(len(self.mock_tracker.tracks), original_count)
        for track in mock_tracks:
            self.assertIn(track, self.mock_tracker.tracks)

    def test_remove_duplicate_tracks_with_duplicates(self):
        """测试存在重复ID的情况"""
        # 创建重复ID的轨迹
        mock_tracks = []
        
        # ID=1的轨迹，time_since_update=2（应被保留）
        track1_keep = Mock()
        track1_keep.id = 1
        track1_keep.time_since_update = 2
        mock_tracks.append(track1_keep)
        
        # ID=1的轨迹，time_since_update=5（应被移除）
        track1_remove = Mock()
        track1_remove.id = 1
        track1_remove.time_since_update = 5
        mock_tracks.append(track1_remove)
        
        # ID=2的轨迹，无重复
        track2_unique = Mock()
        track2_unique.id = 2
        track2_unique.time_since_update = 1
        mock_tracks.append(track2_unique)
        
        # ID=3的轨迹，time_since_update=0（应被保留）
        track3_keep = Mock()
        track3_keep.id = 3
        track3_keep.time_since_update = 0
        mock_tracks.append(track3_keep)
        
        # ID=3的轨迹，time_since_update=3（应被移除）
        track3_remove = Mock()
        track3_remove.id = 3
        track3_remove.time_since_update = 3
        mock_tracks.append(track3_remove)
        
        self.mock_tracker.tracks = mock_tracks.copy()
        
        # 调用方法
        self.strongsort._remove_duplicate_tracks()
        
        # 验证结果
        self.assertEqual(len(self.mock_tracker.tracks), 3)  # 应该剩余3个轨迹
        
        # 验证保留的轨迹
        remaining_tracks = self.mock_tracker.tracks
        self.assertIn(track1_keep, remaining_tracks)
        self.assertIn(track2_unique, remaining_tracks)
        self.assertIn(track3_keep, remaining_tracks)
        
        # 验证被移除的轨迹
        self.assertNotIn(track1_remove, remaining_tracks)
        self.assertNotIn(track3_remove, remaining_tracks)

    def test_remove_duplicate_tracks_same_time_since_update(self):
        """测试重复ID且time_since_update相同的情况"""
        # 创建相同ID和time_since_update的轨迹
        track1 = Mock()
        track1.id = 1
        track1.time_since_update = 2
        
        track2 = Mock()
        track2.id = 1
        track2.time_since_update = 2
        
        self.mock_tracker.tracks = [track1, track2]
        
        # 调用方法
        self.strongsort._remove_duplicate_tracks()
        
        # 验证只保留一个轨迹（min函数会返回第一个遇到的最小值）
        self.assertEqual(len(self.mock_tracker.tracks), 1)
        # 应该保留第一个轨迹
        self.assertIn(track1, self.mock_tracker.tracks)
        self.assertNotIn(track2, self.mock_tracker.tracks)

    def test_remove_duplicate_tracks_empty_tracks(self):
        """测试空轨迹列表的情况"""
        self.mock_tracker.tracks = []
        
        # 调用方法（不应该抛出异常）
        self.strongsort._remove_duplicate_tracks()
        
        # 验证轨迹列表仍为空
        self.assertEqual(len(self.mock_tracker.tracks), 0)

    def test_remove_duplicate_tracks_multiple_groups(self):
        """测试多个重复ID组的情况"""
        mock_tracks = []
        
        # ID=1组：3个轨迹，保留time_since_update=1的
        for i, time_update in enumerate([3, 1, 5]):
            track = Mock()
            track.id = 1
            track.time_since_update = time_update
            mock_tracks.append(track)
        
        # ID=2组：2个轨迹，保留time_since_update=0的
        for i, time_update in enumerate([2, 0]):
            track = Mock()
            track.id = 2
            track.time_since_update = time_update
            mock_tracks.append(track)
        
        # ID=3：单个轨迹，无重复
        track_unique = Mock()
        track_unique.id = 3
        track_unique.time_since_update = 4
        mock_tracks.append(track_unique)
        
        self.mock_tracker.tracks = mock_tracks.copy()
        
        # 调用方法
        self.strongsort._remove_duplicate_tracks()
        
        # 验证结果：应该剩余3个轨迹（每个ID组保留1个）
        self.assertEqual(len(self.mock_tracker.tracks), 3)
        
        # 验证保留的轨迹具有正确的time_since_update值
        remaining_ids_and_times = [(track.id, track.time_since_update) for track in self.mock_tracker.tracks]
        expected_combinations = [(1, 1), (2, 0), (3, 4)]
        
        for expected in expected_combinations:
            self.assertIn(expected, remaining_ids_and_times)

    def test_remove_duplicate_tracks_large_scale(self):
        """测试大规模重复ID的情况"""
        mock_tracks = []
        
        # 创建大量重复ID的轨迹
        for track_id in range(1, 6):  # ID 1-5
            for time_update in range(10):  # 每个ID有10个轨迹
                track = Mock()
                track.id = track_id
                track.time_since_update = time_update
                mock_tracks.append(track)
        
        self.mock_tracker.tracks = mock_tracks.copy()
        
        # 调用方法
        self.strongsort._remove_duplicate_tracks()
        
        # 验证结果：应该剩余5个轨迹（每个ID保留1个）
        self.assertEqual(len(self.mock_tracker.tracks), 5)
        
        # 验证每个ID都保留了time_since_update=0的轨迹
        remaining_ids_and_times = [(track.id, track.time_since_update) for track in self.mock_tracker.tracks]
        expected_combinations = [(i, 0) for i in range(1, 6)]
        
        for expected in expected_combinations:
            self.assertIn(expected, remaining_ids_and_times)

    def test_remove_duplicate_tracks_edge_cases(self):
        """测试边界情况"""
        # 测试负数ID
        track_negative = Mock()
        track_negative.id = -1
        track_negative.time_since_update = 0
        
        # 测试零ID
        track_zero = Mock()
        track_zero.id = 0
        track_zero.time_since_update = 1
        
        # 测试大数ID
        track_large = Mock()
        track_large.id = 999999
        track_large.time_since_update = 2
        
        self.mock_tracker.tracks = [track_negative, track_zero, track_large]
        
        # 调用方法
        self.strongsort._remove_duplicate_tracks()
        
        # 验证所有轨迹都被保留（因为没有重复ID）
        self.assertEqual(len(self.mock_tracker.tracks), 3)
        self.assertIn(track_negative, self.mock_tracker.tracks)
        self.assertIn(track_zero, self.mock_tracker.tracks)
        self.assertIn(track_large, self.mock_tracker.tracks)


if __name__ == '__main__':
    unittest.main(verbosity=2)


# Mikel Broström 🔥 Yolo Tracking 🧾 AGPL-3.0 license

from pathlib import Path

import numpy as np
from torch import device

from boxmot.appearance.reid.auto_backend import ReidAutoBackend
from boxmot.motion.cmc import get_cmc_method
from boxmot.trackers.basetracker import BaseTracker
from boxmot.trackers.strongsort.sort.detection import Detection
from boxmot.trackers.strongsort.sort.tracker import Tracker
from boxmot.utils.matching import NearestNeighborDistanceMetric
from boxmot.utils.ops import xyxy2tlwh


class StrongSort(object):
    """
    StrongSORT Tracker: A tracking algorithm that utilizes a combination of appearance and motion-based tracking.

    Args:
        model_weights (str): Path to the model weights for ReID (Re-Identification).
        device (str): Device on which to run the model (e.g., 'cpu' or 'cuda').
        fp16 (bool): Whether to use half-precision (fp16) for faster inference on compatible devices.
        per_class (bool, optional): Whether to perform per-class tracking. If True, tracks are maintained separately for each object class.
        max_dist (float, optional): Maximum cosine distance for ReID feature matching in Nearest Neighbor Distance Metric.
        max_iou_dist (float, optional): Maximum Intersection over Union (IoU) distance for data association. Controls the maximum allowed distance between tracklets and detections for a match.
        max_age (int, optional): Maximum number of frames to keep a track alive without any detections.
        n_init (int, optional): Number of consecutive frames required to confirm a track.
        nn_budget (int, optional): Maximum size of the feature library for Nearest Neighbor Distance Metric. If the library size exceeds this value, the oldest features are removed.
        mc_lambda (float, optional): Weight for motion consistency in the track state estimation. Higher values give more weight to motion information.
        ema_alpha (float, optional): Alpha value for exponential moving average (EMA) update of appearance features. Controls the contribution of new and old embeddings in the ReID model.
    """

    def __init__(
        self,
        reid_weights: Path,
        device: device,
        half: bool,
        per_class: bool = False,
        min_conf: float = 0.1,
        max_cos_dist=0.2,
        max_iou_dist=0.7,
        max_age=30,
        n_init=3,
        nn_budget=100,
        mc_lambda=0.98,
        ema_alpha=0.9,
    ):

        self.per_class = per_class
        self.min_conf = min_conf
        self.model = ReidAutoBackend(
            weights=reid_weights, device=device, half=half
        ).model

        self.tracker = Tracker(
            metric=NearestNeighborDistanceMetric("cosine", max_cos_dist, nn_budget),
            max_iou_dist=max_iou_dist,
            max_age=max_age,
            n_init=n_init,
            mc_lambda=mc_lambda,
            ema_alpha=ema_alpha,
        )
        self.cmc = get_cmc_method("ecc")()

    def _apply_target_id_mapping(self, target_tracker_id: np.ndarray, detections: list) -> None:
        """
        应用目标ID映射逻辑，将检测到的目标ID映射到对应的轨迹上
        
        Args:
            target_tracker_id (np.ndarray): 检测结果中的目标tracker ID数组，形状为(N,)
            detections (list): 检测对象列表，与target_tracker_id一一对应
            
        Note:
            - 建立检测与轨迹之间的一对一映射关系
            - 只对确认且最近更新的轨迹进行ID映射
            - 遵循单一职责原则，专门处理ID映射逻辑
            - 包含异常处理，确保跟踪系统的稳定性
        """
        if not self._has_valid_target_ids(target_tracker_id):
            return
            
        try:
            # 创建检测索引到目标ID的映射
            detection_id_mapping = self._create_detection_id_mapping(target_tracker_id, detections)
            
            # 根据轨迹关联的检测更新轨迹ID
            self._update_tracks_with_detection_mapping(detection_id_mapping)

            # 处理重复ID的tracks，保留time_since_update最小的记录
            self._remove_duplicate_tracks()
            
        except Exception as e:
            print(f"目标ID映射过程中发生错误: {e}")
            # 异常情况下继续使用原有跟踪逻辑，不影响系统稳定性
    
    def _has_valid_target_ids(self, target_tracker_id: np.ndarray) -> bool:
        """
        验证目标tracker ID数组是否包含有效数据
        
        Args:
            target_tracker_id (np.ndarray): 目标tracker ID数组
            
        Returns:
            bool: 如果包含有效目标ID返回True，否则返回False
        """
        return (
            len(target_tracker_id) > 0 and 
            np.any(target_tracker_id > 0)  # 至少有一个有效的目标ID
        )

    def _create_detection_id_mapping(self, target_tracker_id: np.ndarray, detections: list) -> dict:
        """
        创建检测索引到目标ID的映射字典
        
        Args:
            target_tracker_id (np.ndarray): 目标tracker ID数组
            detections (list): 检测对象列表
            
        Returns:
            dict: 检测索引到目标ID的映射，格式为 {det_ind: target_id}
        """
        detection_id_mapping = {}

        for i, (detection, target_id) in enumerate(zip(detections, target_tracker_id)):
            if target_id > 0:  # 只处理有效的目标ID
                det_ind = detection.det_ind  # 获取检测索引
                detection_id_mapping[det_ind] = int(target_id)
                print(f"添加映射-> target_id:{target_id},det_ind :{detection.det_ind}")
                
        return detection_id_mapping
    
    def _update_tracks_with_detection_mapping(self, detection_id_mapping: dict) -> None:
        """
        根据检测映射更新轨迹ID
        
        Args:
            detection_id_mapping (dict): 检测索引到目标ID的映射
            
        Note:
            - 只更新已确认且最近更新的轨迹
            - 通过轨迹的det_ind属性找到对应的目标ID
            - 记录ID更新过程用于调试
            轨迹ID更新: 1 -> 1 (检测索引: 1.0)
        """
        for track in self.tracker.tracks:
            if self._should_update_track_id(track):
                # 通过轨迹关联的检测索引查找对应的目标ID
                if hasattr(track, 'det_ind') and track.det_ind in detection_id_mapping:
                    original_id = track.id  # 保存原始ID用于调试
                    target_id = detection_id_mapping[track.det_ind]  # 获取对应的目标ID
                    track.id = target_id    # 更新轨迹ID
                    print(f"轨迹ID更新: {original_id} -> {target_id} (检测索引: {track.det_ind}) -> detection_id_mapping:{detection_id_mapping}")
        
    def _remove_duplicate_tracks(self) -> None:
        """
        移除重复ID的tracks，保留time_since_update最小的记录
        
        Note:
            - 识别具有相同ID的tracks
            - 保留time_since_update最小值的track
            - 将其他重复的tracks标记为删除状态
        """
        # 按ID分组tracks
        id_to_tracks = {}
        for track in self.tracker.tracks:
            if track.id not in id_to_tracks:
                id_to_tracks[track.id] = []
            id_to_tracks[track.id].append(track)
        
        # 处理每个ID组中的重复tracks
        tracks_to_remove = []
        for track_id, tracks in id_to_tracks.items():
            if len(tracks) > 1:  # 存在重复ID
                print(f"发现重复ID {track_id}，共{len(tracks)}个tracks")
                
                # 找到time_since_update最小的track
                best_track = min(tracks, key=lambda t: t.time_since_update)
                print(f"保留track ID={best_track.id}, time_since_update={best_track.time_since_update}")
                
                # 标记其他tracks为删除
                for track in tracks:
                    if track != best_track:
                        print(f"标记删除track ID={track.id}, time_since_update={track.time_since_update}")
                        tracks_to_remove.append(track)
        
        # 从tracks列表中移除重复的tracks
        for track in tracks_to_remove:
            if track in self.tracker.tracks:
                self.tracker.tracks.remove(track)
                print(f"已移除重复track ID={track.id}")
    
    def _should_update_track_id(self, track) -> bool:
        """
        判断是否应该更新轨迹的ID
        
        Args:
            track: 轨迹对象
            
        Returns:
            bool: 如果应该更新返回True，否则返回False
            
        Note:
            - 只有确认的轨迹才会被更新
        """
        return track.is_confirmed()

    @BaseTracker.per_class_decorator
    def update(self, dets: np.ndarray, img: np.ndarray, embs: np.ndarray = None) -> np.ndarray:
        print(f"------ dets: {dets}")

        assert isinstance(
            dets, np.ndarray
        ), f"Unsupported 'dets' input format '{type(dets)}', valid format is np.ndarray"
        assert isinstance(
            img, np.ndarray
        ), f"Unsupported 'img' input format '{type(img)}', valid format is np.ndarray"
        assert (
            len(dets.shape) == 2
        ), "Unsupported 'dets' dimensions, valid number of dimensions is two"
        assert (
            dets.shape[1] == 8
        ), "Unsupported 'dets' 2nd dimension lenght, valid lenghts is 8"

        dets = np.hstack([dets, np.arange(len(dets)).reshape(-1, 1)])
        remain_inds = dets[:, 4] >= self.min_conf
        dets = dets[remain_inds]

        xyxy = dets[:, 0:4]
        confs = dets[:, 4]
        clss = dets[:, 5]
        target_tracker_id = dets[:, 6] # 新增：目标trackerID
        det_ind = dets[:, 7]

        if len(self.tracker.tracks) >= 1:
            warp_matrix = self.cmc.apply(img, xyxy)
            for track in self.tracker.tracks:
                track.camera_update(warp_matrix)

        # extract appearance information for each detection
        if embs is not None:
            features = embs[remain_inds]
        else:
            features = self.model.get_features(xyxy, img)

        tlwh = xyxy2tlwh(xyxy)
        detections = [
            Detection(box, conf, cls, det_ind, feat)
            for box, conf, cls, det_ind, feat in zip(
                tlwh, confs, clss, det_ind, features
            )
        ]
        # 打印detections的详细信息
        print(f"---- 即将进行 update 的 detections ({len(detections)}个):")
        for i, det in enumerate(detections):
            print(f"  Detection[{i}]: tlwh={det.tlwh}, conf={det.conf:.3f}, cls={det.cls}, det_ind={det.det_ind}, feat_shape={det.feat.shape if det.feat is not None else None}")
        
        # 打印tracks的详细信息
        print(f"---- before update tracks len: {len(self.tracker.tracks)}; tracks详情:")
        for i, track in enumerate(self.tracker.tracks):
            bbox = track.to_tlbr()
            state_name = {1: 'Tentative', 2: 'Confirmed', 3: 'Deleted'}.get(track.state, 'Unknown')
            print(f"  Track[{i}]: id={track.id}, bbox=[{bbox[0]:.1f},{bbox[1]:.1f},{bbox[2]:.1f},{bbox[3]:.1f}], conf={track.conf:.3f}, cls={track.cls}, det_ind={track.det_ind}, state={state_name}, time_since_update={track.time_since_update}, hits={track.hits}")

        # update tracker
        self.tracker.predict()
        self.tracker.update(detections)
        # 打印update后tracks的详细信息
        print(f"---- after update tracks len: {len(self.tracker.tracks)}; tracks详情:")
        for i, track in enumerate(self.tracker.tracks):
            bbox = track.to_tlbr()
            state_name = {1: 'Tentative', 2: 'Confirmed', 3: 'Deleted'}.get(track.state, 'Unknown')
            print(f"  Track[{i}]: id={track.id}, bbox=[{bbox[0]:.1f},{bbox[1]:.1f},{bbox[2]:.1f},{bbox[3]:.1f}], conf={track.conf:.3f}, cls={track.cls}, det_ind={track.det_ind}, state={state_name}, time_since_update={track.time_since_update}, hits={track.hits}")

        # 应用目标ID映射逻辑
        self._apply_target_id_mapping(target_tracker_id, detections)

        # output bbox identities
        outputs = []
        for track in self.tracker.tracks:
            if not track.is_confirmed() or track.time_since_update >= 1:
                print(f"------ skip this track, biz track {track.is_confirmed()} is not confirmed or track time_since_update {track.time_since_update} < 1")
                continue

            x1, y1, x2, y2 = track.to_tlbr()

            id = track.id
            conf = track.conf
            cls = track.cls
            det_ind = track.det_ind

            outputs.append(
                np.array([*track.to_tlbr(), id, conf, cls, det_ind]).reshape(1, -1)
            )
        if len(outputs) > 0:
            print(f"------ outputs: {outputs}")
            return np.concatenate(outputs)
        return np.array([])



# Mikel Broström 🔥 Yolo Tracking 🧾 AGPL-3.0 license

import numpy as np
import torch
import time
import pandas as pd
from ultralytics.utils import ops
from ultralytics.engine.results import Results
from typing import Union
from pathlib import Path
import json
import shutil
import os
import sys
import git
import requests
import zipfile
import subprocess
from git import Repo, exc
from boxmot.utils import logger as LOGGER
from tqdm import tqdm
from boxmot.utils import EXAMPLES, ROOT


def split_dataset(src_fldr: Path, percent_to_delete: float = 0.5) -> None:
    """
    Copies the dataset to a new location and removes a specified percentage of images and annotations,
    adjusting the frame index to start at 1.

    Args:
        src_fldr (Path): Source folder containing the dataset.
        percent_to_delete (float): Percentage of images and annotations to remove.
    """
    # Ensure source path is a Path object
    src_fldr = Path(src_fldr)

    # Generate the destination path by replacing "MOT17" with "MOT17-half" in the source path
    new_benchmark_name = f'MOT17-{int(percent_to_delete * 100)}'
    dst_fldr = Path(str(src_fldr).replace('MOT17', new_benchmark_name))

    # Copy the dataset to a new location manually using pathlib if it doesn't already exist
    if not dst_fldr.exists():
        dst_fldr.mkdir(parents=True)
        for item in src_fldr.rglob('*'):
            if item.is_dir():
                (dst_fldr / item.relative_to(src_fldr)).mkdir(parents=True, exist_ok=True)
            else:
                (dst_fldr / item.relative_to(src_fldr)).write_bytes(item.read_bytes())

    # List all sequences in the destination folder
    seq_paths = [f for f in dst_fldr.iterdir() if f.is_dir()]

    # Iterate over each sequence and remove a percentage of images and annotations
    for seq_path in seq_paths:
        seq_gt_path = seq_path / 'gt' / 'gt.txt'
        
        # Check if the gt.txt file exists
        if not seq_gt_path.exists():
            print(f"Ground truth file not found for {seq_path}. Skipping...")
            continue

        df = pd.read_csv(seq_gt_path, sep=",", header=None)
        nr_seq_imgs = df[0].unique().max()
        split = int(nr_seq_imgs * (1 - percent_to_delete))
        
        # Check if the sequence is already split
        if nr_seq_imgs <= split:
            print(f'Sequence {seq_path} already split. Skipping...')
            continue
        
        print(f'Number of annotated frames in {seq_path}: Keeping from frame {split + 1} to {nr_seq_imgs}')

        # Keep rows from the ground truth file beyond the split point
        df = df[df[0] > split]
        
        # Adjust the frame indices to start from 1
        df[0] = df[0] - split

        df.to_csv(seq_gt_path, header=None, index=None, sep=',')

        # Remove images before the split point using pathlib
        jpg_folder_path = seq_path / 'img1'
        jpg_paths = list(jpg_folder_path.glob('*.jpg'))
        for jpg_path in jpg_paths:
            # Extract frame number from image file name (e.g., '000300.jpg' -> 300)
            frame_number = int(jpg_path.stem)
            # Check if this frame number is in the removed range
            if frame_number <= split:
                jpg_path.unlink()

        # Rename the remaining images to have a continuous sequence starting from 1
        remaining_jpg_paths = sorted(jpg_folder_path.glob('*.jpg'))
        for new_index, jpg_path in enumerate(remaining_jpg_paths, start=1):
            new_jpg_name = f"{new_index:06}.jpg"  # zero-padded to 6 digits
            jpg_path.rename(jpg_folder_path / new_jpg_name)

        remaining_images = len(list(jpg_folder_path.glob('*.jpg')))
        print(f'Number of images in {seq_path} after delete: {remaining_images}')
        
    return dst_fldr, new_benchmark_name


def download_mot_eval_tools(val_tools_path):
    """
    Download the official evaluation tools for MOT metrics from the GitHub repository.
    
    Parameters:
        val_tools_path (Path): Path to the destination folder where the evaluation tools will be downloaded.
    
    Returns:
        None. Clones the evaluation tools repository and updates deprecated numpy types.
    """
    val_tools_url = "https://github.com/JonathonLuiten/TrackEval"

    try:
        # Clone the repository
        Repo.clone_from(val_tools_url, val_tools_path)
        LOGGER.debug('Official MOT evaluation repo downloaded successfully.')
    except exc.GitError as err:
        LOGGER.debug(f'Evaluation repo already downloaded or an error occurred: {err}')

    # Fix deprecated np.float, np.int & np.bool by replacing them with native Python types
    deprecated_types = {'np.float': 'float', 'np.int': 'int', 'np.bool': 'bool'}
    
    for file_path in val_tools_path.rglob('*'):
        if file_path.suffix in {'.py', '.txt'}:  # only consider .py and .txt files
            try:
                content = file_path.read_text(encoding='utf-8')
                updated_content = content
                for old_type, new_type in deprecated_types.items():
                    updated_content = updated_content.replace(old_type, new_type)
                
                if updated_content != content:  # Only write back if there were changes
                    file_path.write_text(updated_content, encoding='utf-8')
                    LOGGER.info(f'Replaced deprecated types in {file_path}.')
            except Exception as e:
                LOGGER.error(f'Error processing {file_path}: {e}')


def download_mot_dataset(val_tools_path, benchmark, max_retries=5, backoff_factor=2):
    """
    Download a specific MOT dataset zip file with resumable support and retry logic.
    
    Parameters:
        val_tools_path (Path): Path to the destination folder where the MOT benchmark zip will be downloaded.
        benchmark (str): The MOT benchmark to download (e.g., 'MOT20', 'MOT17').
        max_retries (int): Maximum number of retries for the download in case of failure.
        backoff_factor (int): Exponential backoff factor for delays between retries.
    
    Returns:
        Path: The path to the downloaded zip file.
    """
    url = f'https://motchallenge.net/data/{benchmark}.zip'
    zip_dst = val_tools_path / f'{benchmark}.zip'

    retries = 0  # Initialize retry counter

    response = None
    while retries <= max_retries:
        try:
            response = requests.head(url, allow_redirects=True)
            # Consider any status code less than 400 (e.g., 200, 302) as indicating that the resource exists
            if response.status_code < 400:
                # Get the total size of the file from the server
                total_size_in_bytes = int(response.headers.get('content-length', 0))
                
                # Check if there is already a partially or fully downloaded file
                if zip_dst.exists():
                    current_size = zip_dst.stat().st_size
                    
                    # If the file is fully downloaded, skip the download
                    if current_size >= total_size_in_bytes:
                        LOGGER.info(f"{benchmark}.zip is already fully downloaded.")
                        return zip_dst
                    
                    # If the file is partially downloaded, set the range header to resume
                    resume_header = {'Range': f'bytes={current_size}-'}
                    LOGGER.info(f"Resuming download for {benchmark}.zip from byte {current_size}...")
                else:
                    current_size = 0
                    resume_header = {}

                # Start or resume the download
                response = requests.get(url, headers=resume_header, stream=True)
                response.raise_for_status()  # Check for HTTP request errors

                with open(zip_dst, 'ab') as file, tqdm(
                    desc=zip_dst.name,
                    total=total_size_in_bytes,
                    initial=current_size,
                    unit='iB',
                    unit_scale=True,
                    unit_divisor=1024,
                ) as bar:
                    for data in response.iter_content(chunk_size=1024):
                        size = file.write(data)
                        bar.update(size)

                LOGGER.info(f'{benchmark}.zip downloaded successfully.')
                return zip_dst  # If download is successful, return the path

            else:
                LOGGER.warning(f'{benchmark} is not downloadable from {url}')
                return None

        except (requests.HTTPError, requests.ConnectionError) as e:
            if response and response.status_code == 416:  # Handle "Requested Range Not Satisfiable" error
                LOGGER.info(f"{benchmark}.zip is already fully downloaded.")
                return zip_dst
            LOGGER.error(f'Error occurred while downloading {benchmark}.zip: {e}')
            retries += 1
            wait_time = backoff_factor ** retries
            LOGGER.info(f"Retrying download in {wait_time} seconds... (Attempt {retries} of {max_retries})")
            time.sleep(wait_time)  # Exponential backoff delay

        except Exception as e:
            LOGGER.error(f'An unexpected error occurred: {e}')
            retries += 1
            wait_time = backoff_factor ** retries
            LOGGER.info(f"Retrying download in {wait_time} seconds... (Attempt {retries} of {max_retries})")
            time.sleep(wait_time)  # Exponential backoff delay

    LOGGER.error(f"Failed to download {benchmark}.zip after {max_retries} retries.")
    return None


def unzip_mot_dataset(zip_path, val_tools_path, benchmark):
    """
    Unzip a downloaded MOT dataset zip file into the specified directory.
    
    Parameters:
        zip_path (Path): Path to the downloaded MOT benchmark zip file.
        val_tools_path (Path): Base path to the destination folder where the dataset will be unzipped.
        benchmark (str): The MOT benchmark that was downloaded (e.g., 'MOT20', 'MOT17').
    
    Returns:
        None
    """
    if zip_path is None:
        LOGGER.warning(f'No zip file. Skipping unzipping')
        return None

    extract_path = val_tools_path / 'data' / benchmark
    if not extract_path.exists():
        try:
            with zipfile.ZipFile(zip_path, 'r') as zip_ref:
                # folder will be called as the original fetched file
                zip_ref.extractall(val_tools_path / 'data')

            LOGGER.info(f'{benchmark}.zip unzipped successfully.')
        except zipfile.BadZipFile:
            LOGGER.error(f'{zip_path.name} is corrupted. Try deleting the file and run the script again.')
        except Exception as e:
            LOGGER.error(f'An error occurred while unzipping {zip_path.name}: {e}')
    else:
        LOGGER.info(f'{benchmark} folder already exists.')
        return extract_path
    

def set_gt_fps(opt, seq_paths):
    fps_json_filepath = opt.exp_folder_path / 'seqs_frame_nums.json'
    with open(fps_json_filepath, 'r') as f:
        seqs_frame_nums = json.load(f)
    
    for seq_path in seq_paths:
        seq_name = seq_path.parent.name
        frame_nums = seqs_frame_nums[seq_name]

        gt_dir = seq_path.parent / 'gt'
        gt_orig_path = gt_dir / 'gt.txt' 
        gt_temp_path = gt_dir / 'gt_temp.txt' 
        shutil.copy(gt_orig_path, gt_temp_path)

        seq = np.loadtxt(gt_temp_path, delimiter=',')
        seq_filtered = seq[np.isin(seq[:, 0], frame_nums)]
        np.savetxt(gt_temp_path, seq_filtered, delimiter=',')


def eval_setup(opt, val_tools_path):
    """
    Initializes and sets up evaluation paths for MOT challenge datasets.
    
    This function prepares the directories and paths needed for evaluating
    object tracking algorithms on MOT datasets like MOT17 or custom datasets like MOT17-mini.
    It filters sequence paths based on the detector (for MOT17), sets up the ground truth,
    sequences, and results directories according to the provided options.
    
    Parameters:
    - opt: An object with attributes that include benchmark (str), split (str),
      eval_existing (bool), project (str), and name (str). These options dictate
      the dataset to use, the split of the dataset, whether to evaluate on an
      existing setup, and the naming for the project and evaluation results directory.
    - val_tools_path: A string or Path object pointing to the base directory where
      the validation tools and datasets are located.
    
    Returns:
    - seq_paths: A list of Path objects pointing to the sequence directories to be evaluated.
    - save_dir: A Path object pointing to the directory where evaluation results will be saved.
    - MOT_results_folder: A Path object pointing to the directory where MOT challenge
      formatted results should be placed.
    - gt_folder: A Path object pointing to the directory where ground truth data is located.
    """

    # Convert val_tools_path to Path object if it's not already one
    val_tools_path = Path(val_tools_path)
    
    # Initial setup for paths based on benchmark and split options
    mot_seqs_path = val_tools_path / 'data' / opt.benchmark / opt.split
    gt_folder = mot_seqs_path  # Assuming gt_folder is the same as mot_seqs_path initially
    
    # Handling different benchmarks
    if opt.benchmark == 'MOT17':
        # Filter for FRCNN sequences in MOT17
        seq_paths = [p / 'img1' for p in mot_seqs_path.iterdir() if p.is_dir()]
    elif opt.benchmark == 'MOT17-mini':
        # Adjust paths for MOT17-mini
        # 首先检查当前项目目录下是否存在数据集
        project_dataset_path = ROOT / 'assets' / opt.benchmark / opt.split
        if project_dataset_path.exists():
            LOGGER.info(f"Using dataset from project directory: {project_dataset_path}")
            base_path = project_dataset_path
        else:
            # 如果当前项目目录下不存在，则尝试使用用户指定的路径
            user_dataset_path = Path(opt.source).resolve()
            if user_dataset_path.exists():
                LOGGER.info(f"Using dataset from user specified path: {user_dataset_path}")
                base_path = user_dataset_path
            else:
                # 如果用户指定的路径也不存在，则使用默认路径
                LOGGER.info(f"Using default dataset path: {ROOT / 'assets' / opt.benchmark / opt.split}")
                base_path = ROOT / 'assets' / opt.benchmark / opt.split
        mot_seqs_path = gt_folder = base_path
        seq_paths = [p / 'img1' for p in mot_seqs_path.iterdir() if p.is_dir()]
    else:
        # Default handling for other datasets
        seq_paths = [p / 'img1' for p in mot_seqs_path.iterdir() if p.is_dir()]

    # Set FPS for GT files
    set_gt_fps(opt, seq_paths)

    # Determine save directory
    save_dir = Path(opt.project) / opt.name

    # Setup MOT results folder
    MOT_results_folder = val_tools_path / 'data' / 'trackers' / 'mot_challenge' / opt.benchmark / save_dir.name / 'data'
    MOT_results_folder.mkdir(parents=True, exist_ok=True)  # Ensure directory exists
    
    return seq_paths, save_dir, MOT_results_folder, gt_folder


def convert_to_mot_format(results: Union[Results, np.ndarray], frame_idx: int) -> np.ndarray:
    """
    Converts tracking results for a single frame into MOT challenge format.

    This function supports inputs as either a custom object with a 'boxes' attribute or a numpy array.
    For custom object inputs, 'boxes' should contain 'id', 'xyxy', 'conf', and 'cls' sub-attributes.
    For numpy array inputs, the expected format per row is: (xmin, ymin, xmax, ymax, id, conf, cls).

    Parameters:
    - results (Union[Results, np.ndarray]): Tracking results for the current frame.
    - frame_idx (int): The zero-based index of the frame being processed.

    Returns:
    - np.ndarray: An array containing the MOT formatted results for the frame.
    """

    # Check if results are not empty
    if results.size != 0:
        if isinstance(results, np.ndarray):
            # Convert numpy array results to MOT format
            tlwh = ops.xyxy2ltwh(results[:, 0:4])
            frame_idx_column = np.full((results.shape[0], 1), frame_idx, dtype=np.int32)
            mot_results = np.column_stack((
                frame_idx_column, # frame index
                results[:, 4].astype(np.int32),  # track id
                tlwh.round().astype(np.int32),  # top,left,width,height
                np.ones((results.shape[0], 1), dtype=np.int32),  # "not ignored"
                results[:, 6].astype(np.int32),  # class
                results[:, 5],  # confidence (float)
            ))
            return mot_results
        else:
            # Convert ultralytics results to MOT format
            num_detections = len(results.boxes)
            frame_indices = torch.full((num_detections, 1), frame_idx + 1, dtype=torch.int32)
            not_ignored = torch.ones((num_detections, 1), dtype=torch.int32)

            mot_results = torch.cat([
                frame_indices, # frame index
                results.boxes.id.unsqueeze(1).astype(np.int32), # track id
                ops.xyxy2ltwh(results.boxes.xyxy).astype(np.int32),  ## top,left,width,height
                not_ignored, # "not ignored"
                results.boxes.cls.unsqueeze(1).astype(np.int32), # class
                results.boxes.conf.unsqueeze(1).astype(np.float32), # confidence (float)
            ], dim=1)

            return mot_results.numpy()


def write_mot_results(txt_path: Path, mot_results: np.ndarray) -> None:
    """
    Writes the MOT challenge formatted results to a text file.

    Parameters:
    - txt_path (Path): The path to the text file where results are saved.
    - mot_results (np.ndarray): An array containing the MOT formatted results.

    Note: The text file will be created if it does not exist, and the directory
    path to the file will be created as well if necessary.
    """
    if mot_results is not None:
        # Ensure the parent directory of the txt_path exists
        txt_path.parent.mkdir(parents=True, exist_ok=True)

        # Ensure the file exists before opening
        txt_path.touch(exist_ok=True)

        if mot_results.size != 0:
            # Open the file in append mode and save the MOT results
            with open(str(txt_path), 'a') as file:
                np.savetxt(file, mot_results, fmt='%d,%d,%d,%d,%d,%d,%d,%d,%.6f')




# python3 scripts\run_rob_mots.py --ROBMOTS_SPLIT val --TRACKERS_TO_EVAL tracker_name (e.g. STP) --USE_PARALLEL True --NUM_PARALLEL_CORES 4

import sys
import os
import csv
import numpy as np
from multiprocessing import freeze_support

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import trackeval  # noqa: E402
from trackeval import utils
code_path = utils.get_code_path()

if __name__ == '__main__':
    freeze_support()

    script_config = {
        'ROBMOTS_SPLIT': 'train',  # 'train',  # valid: 'train', 'val', 'test', 'test_live', 'test_post', 'test_all'
        'BENCHMARKS': ['kitti_mots', 'davis_unsupervised', 'youtube_vis', 'ovis', 'tao'], # 'bdd_mots' coming soon
        'GT_FOLDER': os.path.join(code_path, 'data/gt/rob_mots'),
        'TRACKERS_FOLDER': os.path.join(code_path, 'data/trackers/rob_mots'),
    }

    default_eval_config = trackeval.Evaluator.get_default_eval_config()
    default_eval_config['PRINT_ONLY_COMBINED'] = True
    default_eval_config['DISPLAY_LESS_PROGRESS'] = True
    default_dataset_config = trackeval.datasets.RobMOTS.get_default_dataset_config()
    config = {**default_eval_config, **default_dataset_config, **script_config}

    # Command line interface:
    config = utils.update_config(config)

    if config['ROBMOTS_SPLIT'] == 'val':
        config['BENCHMARKS'] = ['kitti_mots', 'bdd_mots', 'davis_unsupervised', 'youtube_vis', 'ovis',
                                       'tao', 'mots_challenge']
        config['SPLIT_TO_EVAL'] = 'val'
    elif config['ROBMOTS_SPLIT'] == 'test' or config['SPLIT_TO_EVAL'] == 'test_live':
        config['BENCHMARKS'] = ['kitti_mots', 'bdd_mots', 'davis_unsupervised', 'youtube_vis', 'ovis', 'tao']
        config['SPLIT_TO_EVAL'] = 'test'
    elif config['ROBMOTS_SPLIT'] == 'test_post':
        config['BENCHMARKS'] = ['mots_challenge', 'waymo']
        config['SPLIT_TO_EVAL'] = 'test'
    elif config['ROBMOTS_SPLIT'] == 'test_all':
        config['BENCHMARKS'] = ['kitti_mots', 'bdd_mots', 'davis_unsupervised', 'youtube_vis', 'ovis',
                                       'tao', 'mots_challenge', 'waymo']
        config['SPLIT_TO_EVAL'] = 'test'
    elif config['ROBMOTS_SPLIT'] == 'train':
        config['BENCHMARKS'] = ['kitti_mots', 'davis_unsupervised', 'youtube_vis', 'ovis', 'tao']  # 'bdd_mots' coming soon
        config['SPLIT_TO_EVAL'] = 'train'

    metrics_config = {'METRICS': ['HOTA']}
    # metrics_config = {'METRICS': ['HOTA', 'CLEAR', 'Identity']}
    eval_config = {k: v for k, v in config.items() if k in config.keys()}
    dataset_config = {k: v for k, v in config.items() if k in config.keys()}

    # Run code
    dataset_list = []
    for bench in config['BENCHMARKS']:
        dataset_config['SUB_BENCHMARK'] = bench
        dataset_list.append(trackeval.datasets.RobMOTS(dataset_config))
    evaluator = trackeval.Evaluator(eval_config)
    metrics_list = []
    for metric in [trackeval.metrics.HOTA, trackeval.metrics.CLEAR, trackeval.metrics.Identity]:
        if metric.get_name() in metrics_config['METRICS']:
            metrics_list.append(metric())
    if len(metrics_list) == 0:
        raise Exception('No metrics selected for evaluation')
    output_res, output_msg = evaluator.evaluate(dataset_list, metrics_list)


    # For each benchmark, combine the 'all' score with the 'cls_averaged' using geometric mean.
    metrics_to_calc = ['HOTA', 'DetA', 'AssA', 'DetRe', 'DetPr', 'AssRe', 'AssPr', 'LocA']
    trackers = list(output_res['RobMOTS.' + config['BENCHMARKS'][0]].keys())
    for tracker in trackers:
        # final_results[benchmark][result_type][metric]
        final_results = {}
        res = {bench: output_res['RobMOTS.' + bench][tracker]['COMBINED_SEQ'] for bench in config['BENCHMARKS']}
        for bench in config['BENCHMARKS']:
            final_results[bench] = {'cls_av': {}, 'det_av': {}, 'final': {}}
            for metric in metrics_to_calc:
                final_results[bench]['cls_av'][metric] = np.mean(res[bench]['cls_comb_cls_av']['HOTA'][metric])
                final_results[bench]['det_av'][metric] = np.mean(res[bench]['all']['HOTA'][metric])
                final_results[bench]['final'][metric] = \
                    np.sqrt(final_results[bench]['cls_av'][metric] * final_results[bench]['det_av'][metric])

        # Take the arithmetic mean over all the benchmarks
        final_results['overall'] = {'cls_av': {}, 'det_av': {}, 'final': {}}
        for metric in metrics_to_calc:
            final_results['overall']['cls_av'][metric] = \
                np.mean([final_results[bench]['cls_av'][metric] for bench in config['BENCHMARKS']])
            final_results['overall']['det_av'][metric] = \
                np.mean([final_results[bench]['det_av'][metric] for bench in config['BENCHMARKS']])
            final_results['overall']['final'][metric] = \
                np.mean([final_results[bench]['final'][metric] for bench in config['BENCHMARKS']])

        # Save out result
        headers = [config['SPLIT_TO_EVAL']] + [x + '___' + metric for x in ['f', 'c', 'd'] for metric in metrics_to_calc]

        def rowify(d):
            return [d[x][metric] for x in ['final', 'cls_av', 'det_av'] for metric in metrics_to_calc]

        out_file = os.path.join(script_config['TRACKERS_FOLDER'], script_config['ROBMOTS_SPLIT'], tracker,
                                'final_results.csv')

        with open(out_file, 'w', newline='') as f:
            writer = csv.writer(f, delimiter=',')
            writer.writerow(headers)
            writer.writerow(['overall'] + rowify(final_results['overall']))
            for bench in config['BENCHMARKS']:
                if bench == 'overall':
                    continue
                writer.writerow([bench] + rowify(final_results[bench]))



#!/usr/bin/env python3
"""
BoxMOT包使用示例

本示例展示如何在其他项目中使用boxmot_with_tracker包，
特别是StrongSort追踪器的ID映射功能。

安装包:
    pip install boxmot_with_tracker

或从TestPyPI安装:
    pip install --index-url https://test.pypi.org/simple/ boxmot_with_tracker
"""

import cv2
import numpy as np
from typing import List, Tuple

# 导入BoxMOT
try:
    from boxmot import create_tracker
    print("✅ BoxMOT导入成功")
except ImportError as e:
    print(f"❌ BoxMOT导入失败: {e}")
    print("请先安装包: pip install boxmot_with_tracker")
    exit(1)


class VideoTracker:
    """视频追踪器类
    
    封装BoxMOT追踪器，提供简单易用的接口。
    """
    
    def __init__(self, 
                 tracker_type: str = 'strongsort',
                 reid_weights: str = None,
                 device: str = 'cpu',
                 half: bool = False):
        """初始化追踪器
        
        Args:
            tracker_type: 追踪器类型
            reid_weights: ReID模型权重文件
            device: 计算设备 ('cpu' 或 'cuda')
            half: 是否使用半精度
        """
        # 使用默认配置创建追踪器
        from boxmot.trackers.strongsort.strongsort import StrongSort
        self.tracker = StrongSort(
            reid_weights=reid_weights,
            device=device,
            half=half
        )
        print(f"🎯 追踪器初始化成功: {tracker_type}")
    
    def track_frame(self, 
                   frame: np.ndarray, 
                   detections: np.ndarray) -> np.ndarray:
        """追踪单帧
        
        Args:
            frame: 输入图像 [H, W, C]
            detections: 检测结果 [N, 7] - [x1,y1,x2,y2,conf,class,target_id]
            
        Returns:
            np.ndarray: 追踪结果 [N, 8] - [x1,y1,x2,y2,track_id,conf,class,-1]
        """
        return self.tracker.update(detections, frame)
    
    def reset(self):
        """重置追踪器"""
        # 重新创建追踪器实例
        from boxmot.trackers.strongsort.strongsort import StrongSort
        self.tracker = StrongSort(
            reid_weights=None,
            device='cpu',
            half=False
        )


def create_sample_detections(frame_idx: int) -> np.ndarray:
    """创建示例检测数据
    
    模拟目标检测器的输出，包含目标ID信息。
    
    Args:
        frame_idx: 帧索引
        
    Returns:
        np.ndarray: 检测结果 [N, 7]
    """
    # 模拟移动的目标
    base_x = 100 + frame_idx * 5  # 水平移动
    base_y = 150 + int(10 * np.sin(frame_idx * 0.1))  # 垂直振荡
    
    detections = np.array([
        # [x1, y1, x2, y2, conf, class, target_id]
        [base_x, base_y, base_x + 80, base_y + 120, 0.95, 0, 1001],  # 人员1
        [base_x + 200, base_y + 50, base_x + 280, base_y + 170, 0.88, 0, 1002],  # 人员2
        [base_x + 400, base_y - 20, base_x + 520, base_y + 80, 0.76, 1, 2001],  # 车辆1
    ])
    
    return detections


def draw_tracks(frame: np.ndarray, 
               tracks: np.ndarray, 
               detections: np.ndarray = None) -> np.ndarray:
    """在图像上绘制追踪结果
    
    Args:
        frame: 输入图像
        tracks: 追踪结果
        detections: 原始检测结果（可选）
        
    Returns:
        np.ndarray: 绘制后的图像
    """
    result = frame.copy()
    
    # 定义颜色
    colors = {
        0: (0, 255, 0),    # 人员 - 绿色
        1: (255, 0, 0),    # 车辆 - 蓝色
    }
    
    # 绘制追踪结果
    for track in tracks:
        x1, y1, x2, y2, track_id, conf, cls, _ = track
        x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
        track_id, cls = int(track_id), int(cls)
        
        # 获取颜色
        color = colors.get(cls, (128, 128, 128))
        
        # 绘制边界框
        cv2.rectangle(result, (x1, y1), (x2, y2), color, 2)
        
        # 绘制标签
        label = f"ID:{track_id} Cls:{cls} Conf:{conf:.2f}"
        label_size = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX, 0.5, 1)[0]
        cv2.rectangle(result, (x1, y1 - label_size[1] - 10), 
                     (x1 + label_size[0], y1), color, -1)
        cv2.putText(result, label, (x1, y1 - 5), 
                   cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 1)
    
    # 如果提供了检测结果，也绘制原始目标ID
    if detections is not None:
        for det in detections:
            x1, y1, x2, y2, conf, cls, target_id = det
            x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
            target_id, cls = int(target_id), int(cls)
            
            # 在右上角绘制原始目标ID
            target_label = f"Target:{target_id}"
            cv2.putText(result, target_label, (x2 - 80, y1 + 15), 
                       cv2.FONT_HERSHEY_SIMPLEX, 0.4, (0, 255, 255), 1)
    
    return result


def demo_basic_tracking():
    """基础追踪演示"""
    print("\n=== 基础追踪演示 ===")
    
    # 创建追踪器
    tracker = VideoTracker()
    
    # 创建测试图像
    frame = np.zeros((480, 640, 3), dtype=np.uint8)
    frame.fill(50)  # 深灰色背景
    
    # 模拟多帧追踪
    for frame_idx in range(5):
        print(f"\n📹 处理第 {frame_idx + 1} 帧")
        
        # 生成检测数据
        detections = create_sample_detections(frame_idx)
        print(f"  检测数量: {len(detections)}")
        print(f"  目标ID: {detections[:, 6].astype(int)}")
        
        # 执行追踪
        tracks = tracker.track_frame(frame, detections)
        print(f"  追踪数量: {len(tracks)}")
        
        if len(tracks) > 0:
            track_ids = tracks[:, 4].astype(int)
            print(f"  轨迹ID: {track_ids}")
            
            # 检查ID映射效果
            target_ids = detections[:, 6].astype(int)
            mapped_count = sum(1 for tid in track_ids if tid in target_ids)
            print(f"  ID映射成功: {mapped_count}/{len(target_ids)}")
        else:
            print("  ⚠️  未生成轨迹")


def demo_id_mapping_analysis():
    """ID映射分析演示"""
    print("\n=== ID映射分析演示 ===")
    
    tracker = VideoTracker()
    frame = np.zeros((480, 640, 3), dtype=np.uint8)
    
    # 测试不同的ID映射场景
    test_cases = [
        {
            "name": "正常映射",
            "detections": np.array([
                [100, 100, 200, 200, 0.9, 0, 1001],
                [300, 150, 400, 250, 0.8, 0, 1002],
            ])
        },
        {
            "name": "包含无效ID",
            "detections": np.array([
                [100, 100, 200, 200, 0.9, 0, 1001],
                [300, 150, 400, 250, 0.8, 0, 0],     # 无效ID
                [500, 200, 600, 300, 0.7, 1, -1],    # 无效ID
            ])
        },
        {
            "name": "重复ID",
            "detections": np.array([
                [100, 100, 200, 200, 0.9, 0, 1001],
                [300, 150, 400, 250, 0.8, 0, 1001],  # 重复ID
            ])
        }
    ]
    
    for i, test_case in enumerate(test_cases):
        print(f"\n🧪 测试场景 {i+1}: {test_case['name']}")
        
        detections = test_case['detections']
        target_ids = detections[:, 6].astype(int)
        valid_target_ids = target_ids[target_ids > 0]
        
        print(f"  输入目标ID: {target_ids}")
        print(f"  有效目标ID: {valid_target_ids}")
        
        # 执行追踪
        tracks = tracker.track_frame(frame, detections)
        
        if len(tracks) > 0:
            track_ids = tracks[:, 4].astype(int)
            print(f"  输出轨迹ID: {track_ids}")
            
            # 分析映射效果
            successful_mappings = [tid for tid in track_ids if tid in valid_target_ids]
            print(f"  成功映射: {successful_mappings}")
            print(f"  映射率: {len(successful_mappings)}/{len(valid_target_ids)} = {len(successful_mappings)/len(valid_target_ids)*100:.1f}%" if len(valid_target_ids) > 0 else "  映射率: N/A")
        else:
            print("  ⚠️  未生成轨迹")
        
        # 重置追踪器以避免历史影响
        tracker.reset()


def demo_video_simulation():
    """视频追踪模拟演示"""
    print("\n=== 视频追踪模拟演示 ===")
    
    tracker = VideoTracker()
    
    # 模拟视频参数
    width, height = 640, 480
    num_frames = 10
    
    print(f"📹 模拟视频: {width}x{height}, {num_frames}帧")
    
    # 统计信息
    total_detections = 0
    total_tracks = 0
    successful_mappings = 0
    
    for frame_idx in range(num_frames):
        # 创建帧
        frame = np.random.randint(0, 50, (height, width, 3), dtype=np.uint8)
        
        # 生成检测
        detections = create_sample_detections(frame_idx)
        
        # 执行追踪
        tracks = tracker.track_frame(frame, detections)
        
        # 统计
        total_detections += len(detections)
        total_tracks += len(tracks)
        
        if len(tracks) > 0:
            target_ids = detections[:, 6].astype(int)
            track_ids = tracks[:, 4].astype(int)
            frame_mappings = sum(1 for tid in track_ids if tid in target_ids)
            successful_mappings += frame_mappings
            
            print(f"  帧 {frame_idx+1:2d}: 检测={len(detections)}, 轨迹={len(tracks)}, 映射={frame_mappings}")
        else:
            print(f"  帧 {frame_idx+1:2d}: 检测={len(detections)}, 轨迹=0, 映射=0")
    
    # 输出统计结果
    print(f"\n📊 统计结果:")
    print(f"  总检测数: {total_detections}")
    print(f"  总轨迹数: {total_tracks}")
    print(f"  成功映射: {successful_mappings}")
    print(f"  映射成功率: {successful_mappings/total_detections*100:.1f}%" if total_detections > 0 else "  映射成功率: N/A")


def demo_performance_test():
    """性能测试演示"""
    print("\n=== 性能测试演示 ===")
    
    import time
    
    tracker = VideoTracker()
    frame = np.random.randint(0, 255, (480, 640, 3), dtype=np.uint8)
    
    # 不同规模的检测数据
    test_sizes = [1, 5, 10, 20]
    
    for size in test_sizes:
        print(f"\n🚀 测试规模: {size} 个检测")
        
        # 生成测试数据
        detections = np.random.rand(size, 7)
        detections[:, :4] *= [600, 400, 600, 400]  # 坐标范围
        detections[:, 4] = np.random.uniform(0.5, 1.0, size)  # 置信度
        detections[:, 5] = np.random.randint(0, 2, size)  # 类别
        detections[:, 6] = np.arange(1001, 1001 + size)  # 目标ID
        
        # 性能测试
        times = []
        for _ in range(50):  # 运行50次取平均
            start_time = time.time()
            tracks = tracker.track_frame(frame, detections)
            end_time = time.time()
            times.append(end_time - start_time)
        
        avg_time = np.mean(times) * 1000  # 转换为毫秒
        std_time = np.std(times) * 1000
        fps = 1000 / avg_time if avg_time > 0 else 0
        
        print(f"  平均耗时: {avg_time:.2f} ± {std_time:.2f} ms")
        print(f"  理论FPS: {fps:.1f}")
        print(f"  输出轨迹: {len(tracks) if 'tracks' in locals() else 0}")
        
        # 重置追踪器
        tracker.reset()


def main():
    """主函数"""
    print("🎯 BoxMOT包使用示例")
    print("=" * 50)
    
    try:
        # 运行各种演示
        demo_basic_tracking()
        demo_id_mapping_analysis()
        demo_video_simulation()
        demo_performance_test()
        
        print("\n" + "=" * 50)
        print("🎉 所有演示完成！")
        print("\n📚 更多信息:")
        print("  - 项目文档: README.md")
        print("  - 调试指南: ID_MAPPING_DEBUG_GUIDE.md")
        print("  - 构建指南: PACKAGE_BUILD_GUIDE.md")
        
    except Exception as e:
        print(f"\n❌ 演示过程中出现错误: {e}")
        import traceback
        traceback.print_exc()


if __name__ == "__main__":
    main()